import sys
import warnings

import numpy as np

from pyfits.column import Column, ColDefs, _FormatP, _makep
from pyfits.fitsrec import FITS_rec
from pyfits.hdu.base import DELAYED, ExtensionHDU
from pyfits.hdu.image import _ImageBaseHDU, ImageHDU
from pyfits.hdu.table import BinTableHDU
from pyfits.header import Header
from pyfits.util import (lazyproperty, _pad_length, _is_pseudo_unsigned,
                         _unsigned_zero)

try:
    from pyfits import compression
    COMPRESSION_SUPPORTED = COMPRESSION_ENABLED = True
except ImportError:
    COMPRESSION_SUPPORTED = COMPRESSION_ENABLED = False


# Default compression parameter values

DEFAULT_COMPRESSION_TYPE = 'RICE_1'
DEFAULT_QUANTIZE_LEVEL = 16.
DEFAULT_HCOMP_SCALE = 0
DEFAULT_HCOMP_SMOOTH = 0
DEFAULT_BLOCK_SIZE = 32
DEFAULT_BYTE_PIX = 4


class CompImageHeader(Header):
    """
    Header object for compressed image HDUs designed to keep the compression
    header and the underlying image header properly synchronized.

    This essentially wraps the image header, so that all values are read from
    and written to the image header.  However, updates to the image header will
    also update the table header where appropriate.
    """

    def __init__(self, table_header, image_header=None):
        if image_header is None:
            image_header = Header()
        self._cards = image_header._cards
        self._keyword_indices = image_header._keyword_indices
        self._modified = image_header._modified
        self._table_header = table_header

    def set(self, keyword, value=None, comment=None, before=None, after=None):
        super(CompImageHeader, self).set(keyword, value, comment, before,
                                         after)

        # update the underlying header (_table_header) unless the update
        # was made to a card that describes the data.

        if (keyword not in ('SIMPLE', 'XTENSION', 'BITPIX', 'PCOUNT', 'GCOUNT',
                            'TFIELDS', 'EXTEND', 'ZIMAGE', 'ZBITPIX',
                            'ZCMPTYPE') and
            keyword[:4] not in ('ZVAL') and
            keyword[:5] not in ('NAXIS', 'TTYPE', 'TFORM', 'ZTILE', 'ZNAME')
            and keyword[:6] not in ('ZNAXIS')):
            self._table_header.set(keyword, value, comment, before, after)

    def add_history(self, value, before=None, after=None):
        super(CompImageHeader, self).add_history(value, before, after)
        self._table_header.add_history(value, before, after)

    def add_comment(self, value, before=None, after=None):
        super(CompImageHeader, self).add_comment(value, before, after)
        self._table_header.add_comment(value, before, after)

    def add_blank(self, value='', before=None, after=None):
        super(CompImageHeader, self).add_blank(value, before, after)
        self._table_header.add_blank(value, before, after)


class CompImageHDU(BinTableHDU):
    """
    Compressed Image HDU class.
    """

    def __init__(self, data=None, header=None, name=None,
                 compressionType=DEFAULT_COMPRESSION_TYPE,
                 tileSize=None,
                 hcompScale=DEFAULT_HCOMP_SCALE,
                 hcompSmooth=DEFAULT_HCOMP_SMOOTH,
                 quantizeLevel=DEFAULT_QUANTIZE_LEVEL,
                 do_not_scale_image_data=False,
                 uint=False, scale_back=False, **kwargs):
        """
        Parameters
        ----------
        data : array, optional
            data of the image

        header : Header instance, optional
            header to be associated with the image; when reading the HDU from a
            file (data=DELAYED), the header read from the file

        name : str, optional
            the ``EXTNAME`` value; if this value is `None`, then the name from
            the input image header will be used; if there is no name in the
            input image header then the default name ``COMPRESSED_IMAGE`` is
            used.

        compressionType : str, optional
            compression algorithm 'RICE_1', 'PLIO_1', 'GZIP_1', 'HCOMPRESS_1'

        tileSize : int, optional
            compression tile sizes.  Default treats each row of image as a
            tile.

        hcompScale : float, optional
            HCOMPRESS scale parameter

        hcompSmooth : float, optional
            HCOMPRESS smooth parameter

        quantizeLevel : float, optional
            floating point quantization level; see note below

        Notes
        -----
        The pyfits module supports 2 methods of image compression.

            1) The entire FITS file may be externally compressed with the gzip
               or pkzip utility programs, producing a ``*.gz`` or ``*.zip``
               file, respectively.  When reading compressed files of this type,
               pyfits first uncompresses the entire file into a temporary file
               before performing the requested read operations.
               The pyfits module does not support writing to these
               types of compressed files.  This type of compression is
               supported in the `_File` class, not in the `CompImageHDU` class.
               The file compression type is recognized by the ``.gz`` or
               ``.zip`` file name extension.

            2) The `CompImageHDU` class supports the FITS tiled image
               compression convention in which the image is subdivided into a
               grid of rectangular tiles, and each tile of pixels is
               individually compressed.
               The details of this FITS compression convention are described at
               the `FITS Support Office web site
               <http://fits.gsfc.nasa.gov/registry/tilecompression.html>`_.
               Basically, the compressed image tiles are stored in rows of a
               variable length arrray column in a FITS binary table.  The
               pyfits module recognizes that this binary table extension
               contains an image and treats it as if it were an image
               extension.  Under this tile-compression format, FITS header
               keywords remain uncompressed.  At this time, pyfits does not
               support the ability to extract and uncompress sections of the
               image without having to uncompress the entire image.

        The `pyfits` module supports 3 general-purpose compression algorithms
        plus one other special-purpose compression technique that is designed
        for data masks with positive integer pixel values.  The 3 general
        purpose algorithms are GZIP, Rice, and HCOMPRESS, and the
        special-purpose technique is the IRAF pixel list compression technique
        (PLIO).  The `compressionType` parameter defines the compression
        algorithm to be used.

        The FITS image can be subdivided into any desired rectangular grid of
        compression tiles.  With the GZIP, Rice, and PLIO algorithms, the
        default is to take each row of the image as a tile.  The HCOMPRESS
        algorithm is inherently 2-dimensional in nature, so the default in this
        case is to take 16 rows of the image per tile.  In most cases, it makes
        little difference what tiling pattern is used, so the default tiles are
        usually adequate.  In the case of very small images, it could be more
        efficient to compress the whole image as a single tile.  Note that the
        image dimensions are not required to be an integer multiple of the tile
        dimensions; if not, then the tiles at the edges of the image will be
        smaller than the other tiles.  The `tileSize` parameter may be provided
        as a list of tile sizes, one for each dimension in the image.  For
        example a `tileSize` value of ``[100,100]`` would divide a 300 X 300
        image into 9 100 X 100 tiles.

        The 4 supported image compression algorithms are all 'loss-less' when
        applied to integer FITS images; the pixel values are preserved exactly
        with no loss of information during the compression and uncompression
        process.  In addition, the HCOMPRESS algorithm supports a 'lossy'
        compression mode that will produce larger amount of image compression.
        This is achieved by specifying a non-zero value for the `hcompScale`
        parameter.  Since the amount of compression that is achieved depends
        directly on the RMS noise in the image, it is usually more convenient
        to specify the `hcompScale` factor relative to the RMS noise.  Setting
        `hcompScale` = 2.5 means use a scale factor that is 2.5 times the
        calculated RMS noise in the image tile.  In some cases it may be
        desirable to specify the exact scaling to be used, instead of
        specifying it relative to the calculated noise value.  This may be done
        by specifying the negative of the desired scale value (typically in the
        range -2 to -100).

        Very high compression factors (of 100 or more) can be achieved by using
        large `hcompScale` values, however, this can produce undesireable
        'blocky' artifacts in the compressed image.  A variation of the
        HCOMPRESS algorithm (called HSCOMPRESS) can be used in this case to
        apply a small amount of smoothing of the image when it is uncompressed
        to help cover up these artifacts.  This smoothing is purely cosmetic
        and does not cause any significant change to the image pixel values.
        Setting the `hcompSmooth` parameter to 1 will engage the smoothing
        algorithm.

        Floating point FITS images (which have ``BITPIX`` = -32 or -64) usually
        contain too much 'noise' in the least significant bits of the mantissa
        of the pixel values to be effectively compressed with any lossless
        algorithm.  Consequently, floating point images are first quantized
        into scaled integer pixel values (and thus throwing away much of the
        noise) before being compressed with the specified algorithm (either
        GZIP, RICE, or HCOMPRESS).  This technique produces much higher
        compression factors than simply using the GZIP utility to externally
        compress the whole FITS file, but it also means that the original
        floating point value pixel values are not exactly perserved.  When done
        properly, this integer scaling technique will only discard the
        insignificant noise while still preserving all the real imformation in
        the image.  The amount of precision that is retained in the pixel
        values is controlled by the `quantizeLevel` parameter.  Larger values
        will result in compressed images whose pixels more closely match the
        floating point pixel values, but at the same time the amount of
        compression that is achieved will be reduced.  Users should experiment
        with different values for this parameter to determine the optimal value
        that preserves all the useful information in the image, without
        needlessly preserving all the 'noise' which will hurt the compression
        efficiency.

        The default value for the `quantizeLevel` scale factor is 16, which
        means that scaled integer pixel values will be quantized such that the
        difference between adjacent integer values will be 1/16th of the noise
        level in the image background.  An optimized algorithm is used to
        accurately estimate the noise in the image.  As an example, if the RMS
        noise in the background pixels of an image = 32.0, then the spacing
        between adjacent scaled integer pixel values will equal 2.0 by default.
        Note that the RMS noise is independently calculated for each tile of
        the image, so the resulting integer scaling factor may fluctuate
        slightly for each tile.  In some cases, it may be desireable to specify
        the exact quantization level to be used, instead of specifying it
        relative to the calculated noise value.  This may be done by specifying
        the negative of desired quantization level for the value of
        `quantizeLevel`.  In the previous example, one could specify
        `quantizeLevel`=-2.0 so that the quantized integer levels differ by
        2.0.  Larger negative values for `quantizeLevel` means that the levels
        are more coarsely-spaced, and will produce higher compression factors.
        """

        if not COMPRESSION_SUPPORTED:
            raise Exception('The pyfits.compression module is not available.  '
                            'Creation of compressed image HDUs is disabled.')

        if data is DELAYED:
            # Reading the HDU from a file
            super(CompImageHDU, self).__init__(data=data, header=header)
        else:
            # Create at least a skeleton HDU that matches the input
            # header and data (if any were input)
            super(CompImageHDU, self).__init__(data=None, header=header)

            # Store the input image data
            self.data = data

            # Update the table header (_header) to the compressed
            # image format and to match the input data (if any);
            # Create the image header (_image_header) from the input
            # image header (if any) and ensure it matches the input
            # data; Create the initially empty table data array to
            # hold the compressed data.
            self.updateHeaderData(header, name, compressionType, tileSize,
                                  hcompScale, hcompSmooth, quantizeLevel)

        # TODO: A lot of this should be passed on to an internal image HDU o
        # something like that, see ticket #88
        self._do_not_scale_image_data = do_not_scale_image_data
        self._uint = uint
        self._scale_back = scale_back

        self._axes = [self._header.get('ZNAXIS' + str(axis + 1), 0)
                      for axis in xrange(self._header.get('ZNAXIS', 0))]

        # store any scale factors from the table header
        if do_not_scale_image_data:
            self._bzero = 0
            self._bscale = 1
        else:
            self._bzero = self._header.get('BZERO', 0)
            self._bscale = self._header.get('BSCALE', 1)
        self._bitpix = self._header['ZBITPIX']

        self._orig_bzero = self._bzero
        self._orig_bscale = self._bscale
        self._orig_bitpix = self._bitpix

    @classmethod
    def match_header(cls, header):
        card = header.cards[0]
        if card.keyword != 'XTENSION':
            return False

        xtension = card.value
        if isinstance(xtension, basestring):
            xtension = xtension.rstrip()

        if xtension not in ('BINTABLE', 'A3DTABLE'):
            return False

        if 'ZIMAGE' not in header or header['ZIMAGE'] != True:
            return False

        if COMPRESSION_SUPPORTED and COMPRESSION_ENABLED:
            return True
        elif not COMPRESSION_SUPPORTED:
            warnings.warn('Failure matching header to a compressed image '
                          'HDU: The compression module is not available.\n'
                          'The HDU will be treated as a Binary Table HDU.')
            return False
        else:
            # Compression is supported but disabled; just pass silently (#92)
            return False

    def updateHeaderData(self, image_header,
                         name=None,
                         compressionType=None,
                         tileSize=None,
                         hcompScale=None,
                         hcompSmooth=None,
                         quantizeLevel=None):
        """
        Update the table header (`_header`) to the compressed
        image format and to match the input data (if any).  Create
        the image header (`_image_header`) from the input image
        header (if any) and ensure it matches the input
        data. Create the initially-empty table data array to hold
        the compressed data.

        This method is mainly called internally, but a user may wish to
        call this method after assigning new data to the `CompImageHDU`
        object that is of a different type.

        Parameters
        ----------
        image_header : Header instance
            header to be associated with the image

        name : str, optional
            the ``EXTNAME`` value; if this value is `None`, then the name from
            the input image header will be used; if there is no name in the
            input image header then the default name 'COMPRESSED_IMAGE' is used

        compressionType : str, optional
            compression algorithm 'RICE_1', 'PLIO_1', 'GZIP_1', 'HCOMPRESS_1';
            if this value is `None`, use value already in the header; if no
            value already in the header, use 'RICE_1'

        tileSize : sequence of int, optional
            compression tile sizes as a list; if this value is `None`, use
            value already in the header; if no value already in the header,
            treat each row of image as a tile

        hcompScale : float, optional
            HCOMPRESS scale parameter; if this value is `None`, use the value
            already in the header; if no value already in the header, use 1

        hcompSmooth : float, optional
            HCOMPRESS smooth parameter; if this value is `None`, use the value
            already in the header; if no value already in the header, use 0

        quantizeLevel : float, optional
            floating point quantization level; if this value is `None`, use the
            value already in the header; if no value already in header, use 16
        """

        image_hdu = ImageHDU(data=self.data, header=self._header)
        self._image_header = CompImageHeader(self._header, image_hdu.header)
        self._axes = image_hdu._axes
        del image_hdu

        # Update the extension name in the table header
        if not name and not 'EXTNAME' in self._header:
            name = 'COMPRESSED_IMAGE'

        if name:
            self._header.set('EXTNAME', name,
                             'name of this binary table extension',
                             after='TFIELDS')
            self.name = name
        else:
            self.name = self._header['EXTNAME']

        # Set the compression type in the table header.

        if compressionType:
            if compressionType not in ['RICE_1', 'GZIP_1', 'PLIO_1',
                                       'HCOMPRESS_1']:
                warnings.warn('Unknown compression type provided.  Default '
                              '(%s) compression used.' %
                              DEFAULT_COMPRESSION_TYPE)
                compressionType = DEFAULT_COMPRESSION_TYPE

            self._header.set('ZCMPTYPE', compressionType,
                             'compression algorithm', after='TFIELDS')
        else:
            compressionType = self._header.get('ZCMPTYPE', 'RICE_1')

        # If the input image header had BSCALE/BZERO cards, then insert
        # them in the table header.

        if image_header:
            bzero = image_header.get('BZERO', 0.0)
            bscale = image_header.get('BSCALE', 1.0)
            afterCard = 'EXTNAME'

            if bscale != 1.0:
                self._header.set('BSCALE', bscale, after=afterCard)
                afterCard = 'BSCALE'

            if bzero != 0.0:
                self._header.set('BZERO', bzero, after=afterCard)

            bitpix_comment = image_header.comments['BITPIX']
            naxis_comment = image_header.comments['NAXIS']
        else:
            bitpix_comment = 'data type of original image'
            naxis_comment = 'dimension of original image'

        # Set the label for the first column in the table

        self._header.set('TTYPE1', 'COMPRESSED_DATA', 'label for field 1',
                         after='TFIELDS')

        # Set the data format for the first column.  It is dependent
        # on the requested compression type.

        if compressionType == 'PLIO_1':
            tform1 = '1PI'
        else:
            tform1 = '1PB'

        self._header.set('TFORM1', tform1,
                         'data format of field: variable length array',
                         after='TTYPE1')

        # Create the first column for the table.  This column holds the
        # compressed data.
        col1 = Column(name=self._header['TTYPE1'], format=tform1)

        # Create the additional columns required for floating point
        # data and calculate the width of the output table.

        if self._image_header['BITPIX'] < 0:
            # floating point image has 'COMPRESSED_DATA',
            # 'UNCOMPRESSED_DATA', 'ZSCALE', and 'ZZERO' columns.
            ncols = 4

            # Set up the second column for the table that will hold
            # any uncompressable data.
            self._header.set('TTYPE2', 'UNCOMPRESSED_DATA',
                             'label for field 2', after='TFORM1')

            if self._image_header['BITPIX'] == -32:
                tform2 = '1PE'
            else:
                tform2 = '1PD'

            self._header.set('TFORM2', tform2,
                             'data format of field: variable length array',
                             after='TTYPE2')
            col2 = Column(name=self._header['TTYPE2'], format=tform2)

            # Set up the third column for the table that will hold
            # the scale values for quantized data.
            self._header.set('TTYPE3', 'ZSCALE', 'label for field 3',
                             after='TFORM2')
            self._header.set('TFORM3', '1D',
                             'data format of field: 8-byte DOUBLE',
                             after='TTYPE3')
            col3 = Column(name=self._header['TTYPE3'],
                          format=self._header['TFORM3'])

            # Set up the fourth column for the table that will hold
            # the zero values for the quantized data.
            self._header.set('TTYPE4', 'ZZERO', 'label for field 4',
                             after='TFORM3')
            self._header.set('TFORM4', '1D',
                             'data format of field: 8-byte DOUBLE',
                             after='TTYPE4')
            after = 'TFORM4'
            col4 = Column(name=self._header['TTYPE4'],
                          format=self._header['TFORM4'])

            # Create the ColDefs object for the table
            cols = ColDefs([col1, col2, col3, col4])
        else:
            # default table has just one 'COMPRESSED_DATA' column
            ncols = 1
            after = 'TFORM1'

            # remove any header cards for the additional columns that
            # may be left over from the previous data
            keyList = ['TTYPE2', 'TFORM2', 'TTYPE3', 'TFORM3', 'TTYPE4',
                       'TFORM4']

            for k in keyList:
                try:
                    del self._header[k]
                except KeyError:
                    pass

            # Create the ColDefs object for the table
            cols = ColDefs([col1])

        # Update the table header with the width of the table, the
        # number of fields in the table, the indicator for a compressed
        # image HDU, the data type of the image data and the number of
        # dimensions in the image data array.
        self._header.set('NAXIS1', ncols * 8, 'width of table in bytes')
        self._header.set('TFIELDS', ncols, 'number of fields in each row')
        self._header.set('ZIMAGE', True, 'extension contains compressed image',
                         after=after)
        self._header.set('ZBITPIX', self._image_header['BITPIX'],
                         bitpix_comment, after='ZIMAGE')
        self._header.set('ZNAXIS', self._image_header['NAXIS'], naxis_comment,
                         after='ZBITPIX')

        # Strip the table header of all the ZNAZISn and ZTILEn keywords
        # that may be left over from the previous data

        idx = 1
        while True:
            try:
                del self._header['ZNAXIS' + str(idx)]
                del self._header['ZTILE' + str(idx)]
                idx += 1
            except KeyError:
                break

        # Verify that any input tile size parameter is the appropriate
        # size to match the HDU's data.

        naxis = self._image_header['NAXIS']

        if not tileSize:
            tileSize = []
        elif len(tileSize) != naxis:
            warnings.warn('Provided tile size not appropriate for the data.  '
                          'Default tile size will be used.')
            tileSize = []

        # Set default tile dimensions for HCOMPRESS_1

        if compressionType == 'HCOMPRESS_1':
            if (self._image_header['NAXIS1'] < 4 or
                self._image_header['NAXIS2'] < 4):
                raise ValueError('Hcompress minimum image dimension is '
                                 '4 pixels')
            elif tileSize:
                if tileSize[0] < 4 or tileSize[1] < 4:
                    # user specified tile size is too small
                    raise ValueError('Hcompress minimum tile dimension is '
                                     '4 pixels')
                major_dims = len(filter(lambda x: x > 1, tileSize))
                if major_dims > 2:
                    raise ValueError(
                        'HCOMPRESS can only support 2-dimensional tile sizes.'
                        'All but two of the tileSize dimensions must be set '
                        'to 1.')

            if tileSize and (tileSize[0] == 0 and tileSize[1] == 0):
                #compress the whole image as a single tile
                tileSize[0] = self._image_header['NAXIS1']
                tileSize[1] = self._image_header['NAXIS2']

                for i in range(2, naxis):
                    # set all higher tile dimensions = 1
                    tileSize[i] = 1
            elif not tileSize:
                # The Hcompress algorithm is inherently 2D in nature, so the
                # row by row tiling that is used for other compression
                # algorithms is not appropriate.  If the image has less than 30
                # rows, then the entire image will be compressed as a single
                # tile.  Otherwise the tiles will consist of 16 rows of the
                # image.  This keeps the tiles to a reasonable size, and it
                # also includes enough rows to allow good compression
                # efficiency.  It the last tile of the image happens to contain
                # less than 4 rows, then find another tile size with between 14
                # and 30 rows (preferably even), so that the last tile has at
                # least 4 rows.

                # 1st tile dimension is the row length of the image
                tileSize.append(self._image_header['NAXIS1'])

                if self._image_header['NAXIS2'] <= 30:
                    tileSize.append(self._image_header['NAXIS1'])
                else:
                    # look for another good tile dimension
                    naxis2 = self._image_header['NAXIS2']
                    for dim in [16, 24, 20, 30, 28, 26, 22, 18, 14]:
                        if naxis2 % dim == 0 or naxis2 % dim > 3:
                            tileSize.append(dim)
                            break
                    else:
                        tileSize.append(17)

                for i in range(2, naxis):
                    # set all higher tile dimensions = 1
                    tileSize.append(1)

            # check if requested tile size causes the last tile to have
            # less than 4 pixels

            remain = self._image_header['NAXIS1'] % tileSize[0]  # 1st dimen

            if remain > 0 and remain < 4:
                tileSize[0] += 1  # try increasing tile size by 1

                remain = self._image_header['NAXIS1'] % tileSize[0]

                if remain > 0 and remain < 4:
                    raise ValueError('Last tile along 1st dimension has '
                                     'less than 4 pixels')

            remain = self._image_header['NAXIS2'] % tileSize[1]  # 2nd dimen

            if remain > 0 and remain < 4:
                tileSize[1] += 1  # try increasing tile size by 1

                remain = self._image_header['NAXIS2'] % tileSize[1]

                if remain > 0 and remain < 4:
                    raise ValueError('Last tile along 2nd dimension has '
                                     'less than 4 pixels')

        # Set up locations for writing the next cards in the header.
        last_znaxis = 'ZNAXIS'

        if self._image_header['NAXIS'] > 0:
            after1 = 'ZNAXIS1'
        else:
            after1 = 'ZNAXIS'

        # Calculate the number of rows in the output table and
        # write the ZNAXISn and ZTILEn cards to the table header.
        nrows = 1

        for idx, axis in enumerate(self._axes):
            naxis = 'NAXIS' + str(idx + 1)
            znaxis = 'ZNAXIS' + str(idx + 1)
            ztile = 'ZTILE' + str(idx + 1)

            if tileSize:
                ts = tileSize[idx]
            elif not ztile in self._header:
                # Default tile size
                if not idx:
                    ts = self._image_header['NAXIS1']
                else:
                    ts = 1
            else:
                ts = self._header[ztile]

            nrows = nrows * ((axis - 1) // ts + 1)

            if image_header and naxis in image_header:
                self._header.set(znaxis, axis, image_header.comments[naxis],
                                 after=last_znaxis)
            else:
                self._header.set(znaxis, axis,
                                 'length of original image axis',
                                 after=last_znaxis)

            self._header.set(ztile, ts, 'size of tiles to be compressed',
                             after=after1)
            last_znaxis = znaxis
            after1 = ztile

        # Set the NAXIS2 header card in the table hdu to the number of
        # rows in the table.
        self._header.set('NAXIS2', nrows, 'number of rows in table')

        # Create the record array to be used for the table data.
        self.columns = cols
        compData = np.rec.array(None, formats=','.join(cols._recformats),
                                names=cols.names, shape=nrows)
        self.compData = compData.view(FITS_rec)
        self.compData._coldefs = self.columns
        self.compData.formats = self.columns.formats

        # Set up and initialize the variable length columns.  There will
        # either be one (COMPRESSED_DATA) or two (COMPRESSED_DATA,
        # UNCOMPRESSED_DATA) depending on whether we have floating point
        # data or not.  Note: the ZSCALE and ZZERO columns are fixed
        # length columns.
        for idx in range(min(2, len(cols))):
            self.columns._arrays[idx] = \
                np.rec.recarray.field(self.compData, idx)
            np.rec.recarray.field(self.compData, idx)[0:] = 0
            self.compData._convert[idx] = \
                _makep(self.columns._arrays[idx],
                       np.rec.recarray.field(self.compData, idx),
                       self.columns._recformats[idx])

        # Set the compression parameters in the table header.

        # First, setup the values to be used for the compression parameters
        # in case none were passed in.  This will be either the value
        # already in the table header for that parameter or the default
        # value.
        idx = 1

        while True:
            zname = 'ZNAME' + str(idx)
            if zname not in self._header:
                break
            zval = 'ZVAL' + str(idx)
            if self._header[zname] == 'NOISEBIT':
                if quantizeLevel is None:
                    quantizeLevel = self._header[zval]
            if self._header[zname] == 'SCALE   ':
                if hcompScale is None:
                    hcompScale = self._header[zval]
            if self._header[zname] == 'SMOOTH  ':
                if hcompSmooth is None:
                    hcompSmooth = self._header[zval]
            idx += 1

        if quantizeLevel is None:
            quantizeLevel = DEFAULT_QUANTIZE_LEVEL

        if hcompScale is None:
            hcompScale = DEFAULT_HCOMP_SCALE

        if hcompSmooth is None:
            hcompSmooth = DEFAULT_HCOMP_SCALE

        # Next, strip the table header of all the ZNAMEn and ZVALn keywords
        # that may be left over from the previous data

        idx = 1

        while True:
            zname = 'ZNAME' + str(idx)
            if zname not in self._header:
                break
            zval = 'ZVAL' + str(idx)
            del self._header[zname]
            del self._header[zval]
            idx += 1

        # Finally, put the appropriate keywords back based on the
        # compression type.

        afterCard = 'ZCMPTYPE'
        idx = 1

        if compressionType == 'RICE_1':
            self._header.set('ZNAME1', 'BLOCKSIZE', 'compression block size',
                             after=afterCard)
            self._header.set('ZVAL1', DEFAULT_BLOCK_SIZE, 'pixels per block',
                             after='ZNAME1')

            self._header.set('ZNAME2', 'BYTEPIX',
                             'bytes per pixel (1, 2, 4, or 8)', after='ZVAL1')

            if self._header['ZBITPIX'] == 8:
                bytepix = 1
            elif self._header['ZBITPIX'] == 16:
                bytepix = 2
            else:
                bytepix = DEFAULT_BYTE_PIX

            self._header.set('ZVAL2', bytepix,
                             'bytes per pixel (1, 2, 4, or 8)', after='ZNAME2')
            afterCard = 'ZVAL2'
            idx = 3
        elif compressionType == 'HCOMPRESS_1':
            self._header.set('ZNAME1', 'SCALE', 'HCOMPRESS scale factor',
                             after=afterCard)
            self._header.set('ZVAL1', hcompScale, 'HCOMPRESS scale factor',
                             after='ZNAME1')
            self._header.set('ZNAME2', 'SMOOTH', 'HCOMPRESS smooth option',
                             after='ZVAL1')
            self._header.set('ZVAL2', hcompSmooth, 'HCOMPRESS smooth option',
                             after='ZNAME2')
            afterCard = 'ZVAL2'
            idx = 3

        if self._image_header['BITPIX'] < 0:   # floating point image
            self._header.set('ZNAME' + str(idx), 'NOISEBIT',
                             'floating point quantization level',
                             after=afterCard)
            self._header.set('ZVAL' + str(idx), quantizeLevel,
                             'floating point quantization level',
                             after='ZNAME' + str(idx))

        if image_header:
            # Move SIMPLE card from the image header to the
            # table header as ZSIMPLE card.

            if 'SIMPLE' in image_header:
                self._header.set('ZSIMPLE', image_header['SIMPLE'],
                                 image_header.comments['SIMPLE'],
                                 before='ZBITPIX')

            # Move EXTEND card from the image header to the
            # table header as ZEXTEND card.

            if 'EXTEND' in image_header:
                self._header.set('ZEXTEND', image_header['EXTEND'],
                                 image_header.comments['EXTEND'])

            # Move BLOCKED card from the image header to the
            # table header as ZBLOCKED card.

            if 'BLOCKED' in image_header:
                self._header.set('ZBLOCKED', image_header['BLOCKED'],
                                 image_header.comments['BLOCKED'])

            # Move XTENSION card from the image header to the
            # table header as ZTENSION card.

            # Since we only handle compressed IMAGEs, ZTENSION should
            # always be IMAGE, even if the caller has passed in a header
            # for some other type of extension.
            if 'XTENSION' in image_header:
                self._header.set('ZTENSION', 'IMAGE',
                                 image_header.comments['XTENSION'],
                                 before='ZBITPIX')

            # Move PCOUNT and GCOUNT cards from image header to the table
            # header as ZPCOUNT and ZGCOUNT cards.

            if 'PCOUNT' in image_header:
                self._header.set('ZPCOUNT', image_header['PCOUNT'],
                                 image_header.comments['PCOUNT'],
                                 after=last_znaxis)

            if 'GCOUNT' in image_header:
                self._header.set('ZGCOUNT', image_header['GCOUNT'],
                                 image_header.comments['GCOUNT'],
                                 after='ZPCOUNT')

            # Move CHECKSUM and DATASUM cards from the image header to the
            # table header as XHECKSUM and XDATASUM cards.

            if 'CHECKSUM' in image_header:
                self._header.set('ZHECKSUM', image_header['CHECKSUM'],
                                 image_header.comments['CHECKSUM'])

            if 'DATASUM' in image_header:
                self._header.set('ZDATASUM', image_header['DATASUM'],
                                 image_header.comments['DATASUM'])
        else:
            # Move XTENSION card from the image header to the
            # table header as ZTENSION card.

            # Since we only handle compressed IMAGEs, ZTENSION should
            # always be IMAGE, even if the caller has passed in a header
            # for some other type of extension.
            if 'XTENSION' in self._image_header:
                self._header.set('ZTENSION', 'IMAGE',
                                 self._image_header.comments['XTENSION'],
                                 before='ZBITPIX')

            # Move PCOUNT and GCOUNT cards from image header to the table
            # header as ZPCOUNT and ZGCOUNT cards.

            if 'PCOUNT' in self._image_header:
                self._header.set('ZPCOUNT', self._image_header['PCOUNT'],
                                 self._image_header.comments['PCOUNT'],
                                 after=last_znaxis)

            if 'GCOUNT' in self._image_header:
                self._header.set('ZGCOUNT', self._image_header['GCOUNT'],
                                 self._image_header.comments['GCOUNT'],
                                 after='ZPCOUNT')

        # When we have an image checksum we need to ensure that the same
        # number of blank cards exist in the table header as there were in
        # the image header.  This allows those blank cards to be carried
        # over to the image header when the hdu is uncompressed.

        if 'ZHECKSUM' in self._header:
            required_blanks = image_header._countblanks()
            image_blanks = self._image_header._countblanks()
            table_blanks = self._header._countblanks()

            for _ in range(required_blanks - image_blanks):
                self._image_header.append()
                table_blanks += 1

            for _ in range(required_blanks - table_blanks):
                self._header.append()

    @lazyproperty
    def data(self):
        # The data attribute is the image data (not the table data).

        # First we will get the table data (the compressed
        # data) from the file, if there is any.
        self.compData = super(BinTableHDU, self).data

        # Now that we have the compressed data, we need to uncompress
        # it into the image data.
        dataList = []
        naxesList = []
        tileSizeList = []
        zvalList = []
        uncompressedDataList = []

        # Set up an array holding the integer value that represents
        # undefined pixels.  This could come from the ZBLANK column
        # from the table, or from the ZBLANK header card (if no
        # ZBLANK column (all null values are the same for each tile)),
        # or from the BLANK header card.
        if not 'ZBLANK' in self.compData.names:
            if 'ZBLANK' in self._header:
                nullDvals = np.array(self._header['ZBLANK'],
                                     dtype='int32')
                cn_zblank = -1  # null value is a constant
            elif 'BLANK' in self._header:
                nullDvals = np.array(self._header['BLANK'],
                                     dtype='int32')
                cn_zblank = -1  # null value is a constant
            else:
                cn_zblank = 0  # no null value given so don't check
                nullDvals = np.array(0, dtype='int32')
        else:
            cn_zblank = 1  # null value supplied as a column
            nullDvals = self.compData.field('ZBLANK')

        # Set up an array holding the linear scale factor values
        # This could come from the ZSCALE column from the table, or
        # from the ZSCALE header card (if no ZSCALE column (all
        # linear scale factor values are the same for each tile)).

        if 'BSCALE' in self._header:
            self._bscale = self._header['BSCALE']
            del self._header['BSCALE']
        else:
            self._bscale = 1.

        if not 'ZSCALE' in self.compData.names:
            if 'ZSCALE' in self._header:
                zScaleVals = np.array(self._header['ZSCALE'],
                                      dtype='float64')
                cn_zscale = -1  # scale value is a constant
            else:
                cn_zscale = 0  # no scale factor given so don't scale
                zScaleVals = np.array(1.0, dtype='float64')
        else:
            cn_zscale = 1  # scale value supplied as a column
            zScaleVals = self.compData.field('ZSCALE')

        # Set up an array holding the zero point offset values
        # This could come from the ZZERO column from the table, or
        # from the ZZERO header card (if no ZZERO column (all
        # zero point offset values are the same for each tile)).

        if 'BZERO' in self._header:
            self._bzero = self._header['BZERO']
            del self._header['BZERO']
        else:
            self._bzero = 0.

        if not 'ZZERO' in self.compData.names:
            if 'ZZERO' in self._header:
                zZeroVals = np.array(self._header['ZZERO'],
                                     dtype='float64')
                cn_zzero = -1  # zero value is a constant
            else:
                cn_zzero = 0  # no zero value given so don't scale
                zZeroVals = np.array(1.0, dtype='float64')
        else:
            cn_zzero = 1  # zero value supplied as a column
            zZeroVals = self.compData.field('ZZERO')

        # Is uncompressed data supplied in a column?
        if not 'UNCOMPRESSED_DATA' in self.compData.names:
            cn_uncompressed = 0  # no uncompressed data supplied
        else:
            cn_uncompressed = 1  # uncompressed data supplied as column

        # Take the compressed data out of the array and put it into
        # a list as character bytes to pass to the decompression
        # routine.
        # TODO: I wonder if maybe a lot of this couldn't be done in C by
        # running directly over the array data and not copying it like this
        for row in self.compData:
            dataList.append(row.field('COMPRESSED_DATA').tostring())

            # If we have a column with uncompressed data then create
            # a list of lists of the data in the coulum.  Each
            # underlying list contains the uncompressed data for a
            # pixel in the tile.  There are one of these lists for
            # each tile in the image.
            if 'UNCOMPRESSED_DATA' in self.compData.names:
                tileUncDataList = []

                for byte in row.field('UNCOMPRESSED_DATA'):
                    tileUncDataList.append(byte)

                uncompressedDataList.append(tileUncDataList)

        # Calculate the total number of elements (pixels) in the
        # resulting image data array.  Create a list of the number
        # of pixels along each axis in the image and a list of the
        # number of pixels along each axis in the compressed tile.
        nelem = 1

        for idx in range(self._header['ZNAXIS']):
            naxesList.append(self._header['ZNAXIS' + str(idx + 1)])
            tileSizeList.append(self._header['ZTILE' + str(idx + 1)])
            nelem = nelem * self._header['ZNAXIS' + str(idx + 1)]

        # Create a list for the compression parameters.  The contents
        # of the list is dependent on the compression type.

        if self._header['ZCMPTYPE'] == 'RICE_1':
            idx = 1
            blockSize = DEFAULT_BLOCK_SIZE
            bytePix = DEFAULT_BYTE_PIX

            while True:
                zname = 'ZNAME' + str(idx)
                if zname not in self._header:
                    break
                zval = 'ZVAL' + str(idx)
                if self._header[zname] == 'BLOCKSIZE':
                    blockSize = self._header[zval]
                if self._header[zname] == 'BYTEPIX':
                    bytePix = self._header[zval]
                idx += 1

            zvalList.append(blockSize)
            zvalList.append(bytePix)
        elif self._header['ZCMPTYPE'] == 'HCOMPRESS_1':
            idx = 1
            hcompSmooth = DEFAULT_HCOMP_SMOOTH

            while True:
                zname = 'ZNAME' + str(idx)
                if zname not in self._header:
                    break
                zval = 'ZVAL' + str(idx)
                if self._header[zname] == 'SMOOTH':
                    hcompSmooth = self._header[zval]
                idx += 1

            zvalList.append(hcompSmooth)

        # Treat the NOISEBIT and SCALE parameters separately because
        # they are floats instead of integers

        quantizeLevel = DEFAULT_QUANTIZE_LEVEL

        if self._header['ZBITPIX'] < 0:
            idx = 1

            while True:
                zname = 'ZNAME' + str(idx)
                if zname not in self._header:
                    break
                zval = 'ZVAL' + str(idx)
                if self._header[zname] == 'NOISEBIT':
                    quantizeLevel = self._header[zval]
                idx += 1

        hcompScale = DEFAULT_HCOMP_SCALE

        if self._header['ZCMPTYPE'] == 'HCOMPRESS_1':
            idx = 1

            while True:
                zname = 'ZNAME' + str(idx)
                if zname not in self._header:
                    break
                zval = 'ZVAL' + str(idx)
                if self._header[zname] == 'SCALE':
                    hcompScale = self._header[zval]
                idx += 1

        # Create an array to hold the decompressed data.
        naxesList.reverse()
        data = np.empty(shape=naxesList,
                   dtype=_ImageBaseHDU.NumCode[self._header['ZBITPIX']])
        naxesList.reverse()

        # Call the C decompression routine to decompress the data.
        # Note that any errors in this routine will raise an
        # exception.
        status = compression.decompressData(dataList,
                                            self._header['ZNAXIS'],
                                            naxesList, tileSizeList,
                                            zScaleVals, cn_zscale,
                                            zZeroVals, cn_zzero,
                                            nullDvals, cn_zblank,
                                            uncompressedDataList,
                                            cn_uncompressed,
                                            quantizeLevel,
                                            hcompScale,
                                            zvalList,
                                            self._header['ZCMPTYPE'],
                                            self._header['ZBITPIX'], 1,
                                            nelem, 0.0, data)

        # Scale the data if necessary
        if (self._orig_bzero != 0 or self._orig_bscale != 1):
            new_dtype = self._dtype_for_bitpix()
            data = np.array(data, dtype=new_dtype)

            if cn_zblank:
                blanks = (data == nullDvals)

            if self._bscale != 1:
                np.multiply(data, self._bscale, data)
            if self._bzero != 0:
                data += self._bzero

            if cn_zblank:
                data = np.where(blanks, np.nan, data)

        # Right out of _ImageBaseHDU.data
        self._update_header_scale_info(data.dtype)

        return data

    @data.setter
    def data(self, data):
        if (data is not None) and (not isinstance(data, np.ndarray) or
             data.dtype.fields is not None):
                raise TypeError('CompImageHDU data has incorrect type:%s; '
                                'dtype.fields = %s' %
                                (type(data), data.dtype.fields))

    @lazyproperty
    def compData(self):
        # In order to create the compressed data we will reference the
        # image data.  Referencing the image data will cause the
        # compressed data to be read from the file.
        data = self.data
        return self.compData

    @property
    def shape(self):
        """
        Shape of the image array--should be equivalent to ``self.data.shape``.
        """

        # Determine from the values read from the header
        return tuple(reversed(self._axes))

    @lazyproperty
    def header(self):
        # The header attribute is the header for the image data.  It
        # is not actually stored in the object dictionary.  Instead,
        # the _image_header is stored.  If the _image_header attribute
        # has already been defined we just return it.  If not, we nust
        # create it from the table header (the _header attribute).
        if hasattr(self, '_image_header'):
            return self._image_header

        # Start with a copy of the table header.
        self._image_header = CompImageHeader(self._header, self._header.copy())

        if 'XTENSION' in self._image_header:
            self._image_header['XTENSION'] = ('IMAGE', 'extension type')

        # Delete cards that are related to the table.  And move
        # the values of those cards that relate to the image from
        # their corresponding table cards.  These include
        # ZBITPIX -> BITPIX, ZNAXIS -> NAXIS, and ZNAXISn -> NAXISn.
        try:
            del self._image_header['ZIMAGE']
        except KeyError:
            pass

        try:
            del self._image_header['ZCMPTYPE']
        except KeyError:
            pass

        try:
            del self._image_header['ZBITPIX']
            _bitpix = self._header['ZBITPIX']
            self._image_header['BITPIX'] = (_bitpix,
                                            self._header.comments['ZBITPIX'])
        except KeyError:
            pass

        try:
            del self._image_header['ZNAXIS']
            self._image_header['NAXIS'] = (self._header['ZNAXIS'],
                                           self._header.comments['ZNAXIS'])

            last_naxis = 'NAXIS'
            for idx in range(self._image_header['NAXIS']):
                znaxis = 'ZNAXIS' + str(idx + 1)
                naxis = znaxis[1:]
                del self._image_header[znaxis]
                self._image_header.set(naxis, self._header[znaxis],
                                       self._header.comments[znaxis],
                                       after=last_naxis)
                last_naxis = naxis

            if last_naxis == 'NAXIS1':
                # There is only one axis in the image data so we
                # need to delete the extra NAXIS2 card.
                del self._image_header['NAXIS2']
        except KeyError:
            pass

        try:
            for idx in range(self._header['ZNAXIS']):
                del self._image_header['ZTILE' + str(idx + 1)]

        except KeyError:
            pass

        try:
            del self._image_header['ZPCOUNT']
            self._image_header.set('PCOUNT', self._header['ZPCOUNT'],
                                   self._header.comments['ZPCOUNT'])
        except KeyError:
            try:
                del self._image_header['PCOUNT']
            except KeyError:
                pass

        try:
            del self._image_header['ZGCOUNT']
            self._image_header.set('GCOUNT', self._header['ZGCOUNT'],
                                   self._header.comments['ZGCOUNT'])
        except KeyError:
            try:
                del self._image_header['GCOUNT']
            except KeyError:
                pass

        # Add the appropriate BSCALE and BZERO keywords if the data is scaled;
        # though these will be removed again as soon as the data is read
        # (unless do_not_scale_image_data=True)
        if 'GCOUNT' in self._image_header:
            after = 'GCOUNT'
        else:
            after = None
        if 'BSCALE' in self._header:
            self._image_header.set('BSCALE', self._header['BSCALE'],
                                   self._header.comments['BSCALE'],
                                   after=after)
            after = 'BSCALE'

        if 'BZERO' in self._header:
            self._image_header.set('BZERO', self._header['BZERO'],
                                   self._header.comments['BZERO'],
                                   after=after)

        try:
            del self._image_header['ZEXTEND']
            self._image_header.set('EXTEND', self._header['ZEXTEND'],
                                   self._header.comments['ZEXTEND'],
                                   after=last_naxis)
        except KeyError:
            pass

        try:
            del self._image_header['ZBLOCKED']
            self._image_header.set('BLOCKED', self._header['ZBLOCKED'],
                                   self._header.comments['ZBLOCKED'])
        except KeyError:
            pass

        try:
            del self._image_header['TFIELDS']

            for idx in range(self._header['TFIELDS']):
                del self._image_header['TFORM' + str(idx + 1)]
                ttype = 'TTYPE' + str(idx + 1)
                if ttype in self._image_header:
                    del self._image_header[ttype]

        except KeyError:
            pass

        idx = 1

        while True:
            try:
                del self._image_header['ZNAME' + str(idx)]
                del self._image_header['ZVAL' + str(idx)]
                idx += 1
            except KeyError:
                break

        # Move the ZHECKSUM and ZDATASUM cards to the image header
        # as CHECKSUM and DATASUM
        try:
            del self._image_header['ZHECKSUM']
            self._image_header.set('CHECKSUM', self._header['ZHECKSUM'],
                                   self._header.comments['ZHECKSUM'])
        except KeyError:
            pass

        try:
            del self._image_header['ZDATASUM']
            self._image_header.set('DATASUM', self._header['ZDATASUM'],
                                   self._header.comments['ZDATASUM'])
        except KeyError:
            pass

        try:
            del self._image_header['ZSIMPLE']
            self._image_header.set('SIMPLE', self._header['ZSIMPLE'],
                                   self._header.comments['ZSIMPLE'],
                                   before=1)
            del self._image_header['XTENSION']
        except KeyError:
            pass

        try:
            del self._image_header['ZTENSION']
            if self._header['ZTENSION'] != 'IMAGE':
                warnings.warn("ZTENSION keyword in compressed "
                              "extension != 'IMAGE'")
            self._image_header.set('XTENSION', 'IMAGE',
                                   self._header.comments['ZTENSION'])
        except KeyError:
            pass

        # Remove the EXTNAME card if the value in the table header
        # is the default value of COMPRESSED_IMAGE.

        if ('EXTNAME' in self._header and
            self._header['EXTNAME'] == 'COMPRESSED_IMAGE'):
            del self._image_header['EXTNAME']

        # Look to see if there are any blank cards in the table
        # header.  If there are, there should be the same number
        # of blank cards in the image header.  Add blank cards to
        # the image header to make it so.
        table_blanks = self._header._countblanks()
        image_blanks = self._image_header._countblanks()

        for _ in range(table_blanks - image_blanks):
            self._image_header.append()

        return self._image_header

    def _summary(self):
        """
        Summarize the HDU: name, dimensions, and formats.
        """
        class_name = self.__class__.__name__

        # if data is touched, use data info.
        if self._data_loaded:
            if self.data is None:
                _shape, _format = (), ''
            else:

                # the shape will be in the order of NAXIS's which is the
                # reverse of the numarray shape
                _shape = list(self.data.shape)
                _format = self.data.dtype.name
                _shape.reverse()
                _shape = tuple(_shape)
                _format = _format[_format.rfind('.') + 1:]

        # if data is not touched yet, use header info.
        else:
            _shape = ()

            for idx in range(self.header['NAXIS']):
                _shape += (self.header['NAXIS' + str(idx + 1)],)

            _format = _ImageBaseHDU.NumCode[self.header['BITPIX']]

        return (self.name, class_name, len(self.header), _shape,
                _format)

    def updateCompressedData(self):
        """
        Compress the image data so that it may be written to a file.
        """

        naxesList = []
        tileSizeList = []
        zvalList = []

        # Check to see that the image_header matches the image data
        image_bitpix = _ImageBaseHDU.ImgCode[self.data.dtype.name]

        if (self.header.get('NAXIS', 0) != len(self.data.shape) or
            self.header.get('BITPIX', 0) != image_bitpix or
            self._header.get('ZNAXIS', 0) != len(self.data.shape) or
            self._header.get('ZBITPIX', 0) != image_bitpix or
            self.shape != self.data.shape):
            self.updateHeaderData(self.header)

        # Create lists to hold the number of pixels along each axis of
        # the image data and the number of pixels in each tile of the
        # compressed image.
        for idx in range(self._header['ZNAXIS']):
            naxesList.append(self._header['ZNAXIS' + str(idx + 1)])
            tileSizeList.append(self._header['ZTILE' + str(idx + 1)])

        # Indicate if the linear scale factor is from a column, a single
        # scale value, or not given.
        if 'ZSCALE' in self.compData.names:
            cn_zscale = 1  # there is a scaled column
        elif 'ZSCALE' in self._header:
            cn_zscale = -1  # scale value is a constant
        else:
            cn_zscale = 0  # no scale value given so don't scale

        # Indicate if the zero point offset value is from a column, a
        # single value, or not given.
        if 'ZZERO' in self.compData.names:
            cn_zzero = 1  # there is a scaled column
        elif 'ZZERO' in self._header:
            cn_zzero = -1  # zero value is a constant
        else:
            cn_zzero = 0  # no zero value given so don't scale

        # Indicate if there is a UNCOMPRESSED_DATA column in the
        # compressed data table.
        if 'UNCOMPRESSED_DATA' in self.compData.names:
            cn_uncompressed = 1  # there is a uncompressed data column
        else:
            cn_uncompressed = 0  # there is no uncompressed data column

        # Create a list for the compression parameters.  The contents
        # of the list is dependent on the compression type.

        # TODO: I'm pretty sure most of this is repeated almost exactly a
        # little bit above; take a closer look at whether we can pare this
        # down a bit.

        if self._header['ZCMPTYPE'] == 'RICE_1':
            idx = 1
            blockSize = DEFAULT_BLOCK_SIZE
            bytePix = DEFAULT_BYTE_PIX

            while True:
                zname = 'ZNAME' + str(idx)
                if zname not in self._header:
                    break
                zval = 'ZVAL' + str(idx)
                if self._header[zname] == 'BLOCKSIZE':
                    blockSize = self._header[zval]
                if self._header[zname] == 'BYTEPIX':
                    bytePix = self._header[zval]
                idx += 1

            zvalList.append(blockSize)
            zvalList.append(bytePix)
        elif self._header['ZCMPTYPE'] == 'HCOMPRESS_1':
            idx = 1
            hcompSmooth = DEFAULT_HCOMP_SMOOTH

            while True:
                zname = 'ZNAME' + str(idx)
                if zname not in self._header:
                    break
                zval = 'ZVAL' + str(idx)
                if self._header[zname] == 'SMOOTH':
                    hcompSmooth = self._header[zval]
                idx += 1

            zvalList.append(hcompSmooth)

        # Treat the NOISEBIT and SCALE parameters separately because
        # they are floats instead of integers

        quantizeLevel = DEFAULT_QUANTIZE_LEVEL

        if self._header['ZBITPIX'] < 0:
            idx = 1

            while True:
                zname = 'ZNAME' + str(idx)
                if zname not in self._header:
                    break
                zval = 'ZVAL' + str(idx)
                if self._header[zname] == 'NOISEBIT':
                    quantizeLevel = self._header[zval]
                idx += 1

        hcompScale = DEFAULT_HCOMP_SCALE

        if self._header['ZCMPTYPE'] == 'HCOMPRESS_1':
            idx = 1

            while True:
                zname = 'ZNAME' + str(idx)
                if zname not in self._header:
                    break
                zval = 'ZVAL' + str(idx)
                if self._header[zname] == 'SCALE':
                    hcompScale = self._header[zval]
                idx += 1

        # Indicate if the null value is a constant or if no null value
        # is provided.
        if 'ZBLANK' in self._header:
            cn_zblank = -1  # null value is a constant
            zblank = self._header['ZBLANK']
        else:
            cn_zblank = 0  # no null value so don't use
            zblank = 0

        if 'BSCALE' in self._header and self.data.dtype.str[1] == 'f':
            # If this is scaled data (ie it has a BSCALE value and it is
            # floating point data) then pass in the BSCALE value so the C
            # code can unscale it before compressing.
            cn_bscale = self._header['BSCALE']
        else:
            cn_bscale = 1.0

        if 'BZERO' in self._header and self.data.dtype.str[1] == 'f':
            cn_bzero = self._header['BZERO']
        else:
            cn_bzero = 0.0

        # put data in machine native byteorder on little endian machines
        # for handing off to the compression code
        if sys.byteorder == 'little':
            swap_types = ('>',)
        else:
            swap_types = ('>', '=')

        # TODO: This is copied right out of _ImageBaseHDU._writedata_internal;
        # it would be cool if we could use an internal ImageHDU and use that to
        # write to a buffer for compression or something. See ticket #88
        # deal with unsigned integer 16, 32 and 64 data
        if _is_pseudo_unsigned(self.data.dtype):
            # Convert the unsigned array to signed
            data = np.array(
                self.data - _unsigned_zero(self.data.dtype),
                dtype='<i%d' % self.data.dtype.itemsize)
            should_swap = False
        else:
            data = self.data
            byteorder = data.dtype.str[0]
            should_swap = (byteorder in swap_types)

        if should_swap:
            data.byteswap(True)

        try:
            # Compress the data.
            status, compDataList, scaleList, zeroList, uncompDataList = \
               compression.compressData(data,
                                        self._header['ZNAXIS'],
                                        naxesList, tileSizeList,
                                        cn_zblank, zblank,
                                        cn_bscale, cn_bzero, cn_zscale,
                                        cn_zzero, cn_uncompressed,
                                        quantizeLevel,
                                        hcompScale,
                                        zvalList,
                                        self._header['ZCMPTYPE'],
                                        self.header['BITPIX'], 1,
                                        data.size)
        finally:
            # if data was byteswapped return it to its original order
            if should_swap:
                data.byteswap(True)

        if status != 0:
            raise RuntimeError('Unable to write compressed image')

        # Convert the compressed data from a list of byte strings to
        # an array and set it in the COMPRESSED_DATA field of the table.
        colDType = 'uint8'

        if self._header['ZCMPTYPE'] == 'PLIO_1':
            colDType = 'i2'

        for idx in xrange(len(compDataList)):
            self.compData[idx].setfield('COMPRESSED_DATA',
                                        np.fromstring(compDataList[idx],
                                                      dtype=colDType))

        # Convert the linear scale factor values from a list to an
        # array and set it in the ZSCALE field of the table.
        if cn_zscale > 0:
            for idx in xrange(len(scaleList)):
                self.compData[idx].setfield('ZSCALE', scaleList[idx])

        # Convert the zero point offset values from a list to an
        # array and set it in the ZZERO field of the table.
        if cn_zzero > 0:
            for idx in xrange(len(zeroList)):
                self.compData[idx].setfield('ZZERO', zeroList[idx])

        # Convert the uncompressed data values from a list to an
        # array and set it in the UNCOMPRESSED_DATA field of the table.
        if cn_uncompressed > 0:
            for idx in xrange(len(uncompDataList)):
                self.compData[idx].setfield('UNCOMPRESSED_DATA',
                                            uncompDataList[idx])

        # Update the table header cards to match the compressed data.
        self.updateHeader()

    def updateHeader(self):
        """
        Update the table header cards to match the compressed data.
        """

        # Get the _heapsize attribute to match the data.
        self.compData._scale_back()

        # Check that TFIELDS and NAXIS2 match the data.
        self._header['TFIELDS'] = self.compData._nfields
        self._header['NAXIS2'] = self.compData.shape[0]

        # Calculate PCOUNT, for variable length tables.
        _tbsize = self._header['NAXIS1'] * self._header['NAXIS2']
        _heapstart = self._header.get('THEAP', _tbsize)
        self.compData._gap = _heapstart - _tbsize
        _pcount = self.compData._heapsize + self.compData._gap

        if _pcount > 0:
            self._header['PCOUNT'] = _pcount

        # Update TFORM for variable length columns.
        for idx in range(self.compData._nfields):
            format = self.compData._coldefs._recformats[idx]
            if isinstance(format, _FormatP):
                _max = self.compData.field(idx).max
                format = _FormatP(format.dtype, repeat=format.repeat, max=_max)
                self._header['TFORM' + str(idx + 1)] = format.tform
        # Insure that for RICE_1 that the BLOCKSIZE and BYTEPIX cards
        # are present and set to the hard coded values used by the
        # compression algorithm.
        if self._header['ZCMPTYPE'] == 'RICE_1':
            self._header.set('ZNAME1', 'BLOCKSIZE', 'compression block size',
                             after='ZCMPTYPE')
            self._header.set('ZVAL1', DEFAULT_BLOCK_SIZE, 'pixels per block',
                             after='ZNAME1')

            self._header.set('ZNAME2', 'BYTEPIX',
                             'bytes per pixel (1, 2, 4, or 8)', after='ZVAL1')

            if self._header['ZBITPIX'] == 8:
                bytepix = 1
            elif self._header['ZBITPIX'] == 16:
                bytepix = 2
            else:
                bytepix = DEFAULT_BYTE_PIX

            self._header.set('ZVAL2', bytepix,
                             'bytes per pixel (1, 2, 4, or 8)', after='ZNAME2')

    def scale(self, type=None, option='old', bscale=1, bzero=0):
        """
        Scale image data by using ``BSCALE`` and ``BZERO``.

        Calling this method will scale `self.data` and update the keywords of
        ``BSCALE`` and ``BZERO`` in `self._header` and `self._image_header`.
        This method should only be used right before writing to the output
        file, as the data will be scaled and is therefore not very usable after
        the call.

        Parameters
        ----------

        type : str, optional
            destination data type, use a string representing a numpy dtype
            name, (e.g. ``'uint8'``, ``'int16'``, ``'float32'`` etc.).  If is
            `None`, use the current data type.

        option : str, optional
            how to scale the data: if ``"old"``, use the original ``BSCALE``
            and ``BZERO`` values when the data was read/created. If
            ``"minmax"``, use the minimum and maximum of the data to scale.
            The option will be overwritten by any user-specified bscale/bzero
            values.

        bscale, bzero : int, optional
            user specified ``BSCALE`` and ``BZERO`` values.
        """

        if self.data is None:
            return

        # Determine the destination (numpy) data type
        if type is None:
            type = _ImageBaseHDU.NumCode[self._bitpix]
        _type = getattr(np, type)

        # Determine how to scale the data
        # bscale and bzero takes priority
        if (bscale != 1 or bzero != 0):
            _scale = bscale
            _zero = bzero
        else:
            if option == 'old':
                _scale = self._orig_bscale
                _zero = self._orig_bzero
            elif option == 'minmax':
                if isinstance(_type, np.floating):
                    _scale = 1
                    _zero = 0
                else:
                    _min = np.minimum.reduce(self.data.flat)
                    _max = np.maximum.reduce(self.data.flat)
                    self.data.shape = dims

                    if _type == np.uint8:  # uint8 case
                        _zero = _min
                        _scale = (_max - _min) / (2. ** 8 - 1)
                    else:
                        _zero = (_max + _min) / 2.

                        # throw away -2^N
                        _scale = (_max - _min) / (2. ** (8 * _type.bytes) - 2)

        # Do the scaling
        if _zero != 0:
            self.data += -_zero
            self.header['BZERO'] = _zero
        else:
            try:
                del self.header['BZERO']
            except KeyError:
                pass

        if _scale != 1:
            self.data /= _scale
            self.header['BSCALE'] = _scale
        else:
            try:
                del self.header['BSCALE']
            except KeyError:
                pass

        if self.data.dtype.type != _type:
            self.data = np.array(np.around(self.data), dtype=_type)  # 0.7.7.1

        # Update the BITPIX Card to match the data
        self._bitpix = _ImageBaseHDU.ImgCode[self.data.dtype.name]
        self._bzero = self.header.get('BZERO', 0)
        self._bscale = self.header.get('BSCALE', 1)
        # Update BITPIX for the image header specificially
        # TODO: Make this more clear by using self._image_header, but only once
        # this has been fixed so that the _image_header attribute is guaranteed
        # to be valid
        self.header['BITPIX'] = self._bitpix

        # Update the table header to match the scaled data
        self.updateHeaderData(self.header)

        # Since the image has been manually scaled, the current
        # bitpix/bzero/bscale now serve as the 'original' scaling of the image,
        # as though the original image has been completely replaced
        self._orig_bitpix = self._bitpix
        self._orig_bzero = self._bzero
        self._orig_bscale = self._bscale

    # TODO: Fix this class so that it doesn't actually inherit from
    # BinTableHDU, but instead has an internal BinTableHDU reference
    def _prewriteto(self, checksum=False, inplace=False):
        if self._scale_back:
            self.scale(_ImageBaseHDU.NumCode[self._orig_bitpix])
        if self._data_loaded and self.data is not None:
            self.updateCompressedData()
        # Doesn't call the super's _prewriteto, since it calls
        # self.data._scale_back(), which is meaningless here.
        return ExtensionHDU._prewriteto(self, checksum=checksum,
                                        inplace=inplace)

    def _writeheader(self, fileobj):
        """
        Bypasses `BinTableHDU._writeheader()` which updates the header with
        metadata about the data that is meaningless here; another reason
        why this class maybe shouldn't inherit directly from BinTableHDU...
        """

        return ExtensionHDU._writeheader(self, fileobj)

    def _writedata_internal(self, fileobj):
        """
        Like the normal `BinTableHDU._writedata_internal`(), but we need to
        make sure the byte swap is done on the compressed data and not the
        image data, which requires a little messing with attributes.
        """

        size = 0

        if self.data is not None:
            imagedata = self.data
            # TODO: Ick; have to assign to __dict__ to bypass _setdata; need to
            # find a way to fix this
            self.__dict__['data'] = self.compData
            #self.data = self.compData
            try:
                size += self._binary_table_byte_swap(fileobj)
            finally:
                self.data = imagedata
            size += self.compData.size * self.compData.itemsize

        return size

    # TODO: This was copied right out of _ImageBaseHDU; get rid of it once we
    # find a way to rewrite this class as either a subclass or wrapper for an
    # ImageHDU
    def _dtype_for_bitpix(self):
        """
        Determine the dtype that the data should be converted to depending on
        the BITPIX value in the header, and possibly on the BSCALE value as
        well.  Returns None if there should not be any change.
        """

        bitpix = self._orig_bitpix
        # Handle possible conversion to uints if enabled
        if self._uint and self._orig_bscale == 1:
            for bits, dtype in ((16, np.dtype('uint16')),
                                (32, np.dtype('uint32')),
                                (64, np.dtype('uint64'))):
                if bitpix == bits and self._orig_bzero == 1 << (bits - 1):
                    return dtype

        if bitpix > 16:  # scale integers to Float64
            return np.dtype('float64')
        elif bitpix > 0:  # scale integers to Float32
            return np.dtype('float32')

    def _update_header_scale_info(self, dtype=None):
        if (not self._do_not_scale_image_data and
            not (self._orig_bzero == 0 and self._orig_bscale == 1)):
            for keyword in ['BSCALE', 'BZERO']:
                try:
                    del self.header[keyword]
                    # Since _update_header_scale_info can, currently, be called
                    # *after* _prewriteto(), replace these with blank cards so
                    # the header size doesn't change
                    self.header.append()
                except KeyError:
                    pass

            if dtype is None:
                dtype = self._dtype_for_bitpix()
            if dtype is not None:
                self.header['BITPIX'] = _ImageBaseHDU.ImgCode[dtype.name]

            self._bzero = 0
            self._bscale = 1
            self._bitpix = self.header['BITPIX']

    def _calculate_datasum(self, blocking):
        """
        Calculate the value for the ``DATASUM`` card in the HDU.
        """

        if self._data_loaded and self.data is not None:
            # We have the data to be used.
            return self._calculate_datasum_from_data(self.compData,
                                                     blocking)
        else:
            # This is the case where the data has not been read from the
            # file yet.  We can handle that in a generic manner so we do
            # it in the base class.  The other possibility is that there
            # is no data at all.  This can also be handled in a generic
            # manner.
            return super(CompImageHDU, self)._calculate_datasum(blocking)



from __future__ import division  # confidence high

import csv
import os
import re
import sys
import textwrap
import warnings

import numpy as np
from numpy import char as chararray

# This module may have many dependencies on pyfits.column, but pyfits.column
# has fewer dependencies overall, so it's easier to keep table/column-related
# utilities in pyfits.column
from pyfits.column import (FITS2NUMPY, KEYWORD_NAMES, KEYWORD_ATTRIBUTES,
                           TDEF_RE, Delayed, Column, ColDefs, _ASCIIColDefs,
                           _FormatX, _FormatP, _wrapx, _makep, _VLF,
                           _parse_tformat, _scalar_to_format, _convert_format,
                           _cmp_recformats)
from pyfits.fitsrec import FITS_rec
from pyfits.hdu.base import DELAYED, _ValidHDU, ExtensionHDU
from pyfits.header import Header
from pyfits.util import (lazyproperty, _is_int, _str_to_num, _pad_length,
                         deprecated)


class FITSTableDumpDialect(csv.excel):
    """
    A CSV dialect for the PyFITS format of ASCII dumps of FITS tables.
    """

    delimiter = ' '
    lineterminator = '\n'
    quotechar = '"'
    quoting = csv.QUOTE_ALL
    skipinitialspace = True


class _TableLikeHDU(_ValidHDU):
    """
    A class for HDUs that have table-like data.  This is used for both
    Binary/ASCII tables as well as Random Access Group HDUs (which are
    otherwise too dissimlary for tables to use _TableBaseHDU directly).
    """

    _data_type = FITS_rec

    @classmethod
    def match_header(cls, header):
        """
        This is an abstract HDU type for HDUs that contain table-like data.
        This is even more abstract than _TableBaseHDU which is specifically for
        the standard ASCII and Binary Table types.
        """

        raise NotImplementedError

    @lazyproperty
    def columns(self):
        # The base class doesn't make any assumptions about where the column
        # definitions come from, so just return an empty ColDefs
        return ColDefs([])

    def _get_tbdata(self):
        """Get the table data from an input HDU object."""

        # TODO: Need to find a way to eliminate the check for phantom columns;
        # this detail really needn't be worried about outside the ColDefs class
        columns = self.columns
        recformats = [f for idx, f in enumerate(columns._recformats)
                      if not columns[idx]._phantom]
        formats = ','.join(recformats)
        names = [n for idx, n in enumerate(columns.names)
                 if not columns[idx]._phantom]
        dtype = np.rec.format_parser(formats, names, None).dtype

        # TODO: Details related to variable length arrays need to be dealt with
        # specifically in the BinTableHDU class, since they're a detail
        # specific to FITS binary tables
        if (_FormatP in [type(r) for r in recformats] and
            self._datSpan > self._theap):
            # We have a heap; include it in the raw_data
            raw_data = self._get_raw_data(self._datSpan, np.byte, self._datLoc)
            data = raw_data[:self._theap].view(dtype=dtype,
                                               type=np.rec.recarray)
        else:
            raw_data = self._get_raw_data(columns._shape, dtype, self._datLoc)
            data = raw_data.view(np.rec.recarray)

        self._init_tbdata(data)
        return data.view(self._data_type)

    def _init_tbdata(self, data):
        columns = self.columns

        data.dtype = data.dtype.newbyteorder('>')

        # pass datLoc, for P format
        data._heapoffset = self._theap
        data._heapsize = self._header['PCOUNT']
        data._file = self._file
        data._buffer = self._buffer
        tbsize = self._header['NAXIS1'] * self._header['NAXIS2']
        data._gap = self._theap - tbsize

        # pass the attributes
        fidx = 0
        for idx in range(len(columns)):
            if not columns[idx]._phantom:
                # get the data for each column object from the rec.recarray
                columns[idx].array = data.field(fidx)
                fidx += 1

        # delete the _arrays attribute so that it is recreated to point to the
        # new data placed in the column object above
        del columns._arrays


class _TableBaseHDU(ExtensionHDU, _TableLikeHDU):
    """
    FITS table extension base HDU class.
    """

    def __init__(self, data=None, header=None, name=None):
        """
        Parameters
        ----------
        header : Header instance
            header to be used

        data : array
            data to be used

        name : str
            name to be populated in ``EXTNAME`` keyword
        """

        super(_TableBaseHDU, self).__init__(data=data, header=header,
                                            name=name)

        if header is not None and not isinstance(header, Header):
            raise ValueError('header must be a Header object.')

        if data is DELAYED:
            # this should never happen
            if header is None:
                raise ValueError('No header to setup HDU.')

            # if the file is read the first time, no need to copy, and keep it
            # unchanged
            else:
                self._header = header
        else:
            # construct a list of cards of minimal header
            cards = [
                ('XTENSION',      '', ''),
                ('BITPIX',         8, 'array data type'),
                ('NAXIS',          2, 'number of array dimensions'),
                ('NAXIS1',         0, 'length of dimension 1'),
                ('NAXIS2',         0, 'length of dimension 2'),
                ('PCOUNT',         0, 'number of group parameters'),
                ('GCOUNT',         1, 'number of groups'),
                ('TFIELDS',        0, 'number of table fields')]

            if header is not None:
                # Make a "copy" (not just a view) of the input header, since it
                # may get modified.  the data is still a "view" (for now)
                hcopy = header.copy(strip=True)
                cards.extend(hcopy.cards)

            self._header = Header(cards)

            if isinstance(data, np.ndarray) and data.dtype.fields is not None:
                if isinstance(data, self._data_type):
                    self.data = data
                else:
                    self.data = data.view(self._data_type)

                self._header['NAXIS1'] = self.data.itemsize
                self._header['NAXIS2'] = self.data.shape[0]
                self._header['TFIELDS'] = len(self.data._coldefs)

                self.columns = self.data._coldefs
                self.update()

                try:
                   # Make the ndarrays in the Column objects of the ColDefs
                   # object of the HDU reference the same ndarray as the HDU's
                   # FITS_rec object.
                    for idx in range(len(self.columns)):
                        self.columns[idx].array = self.data.field(idx)

                    # Delete the _arrays attribute so that it is recreated to
                    # point to the new data placed in the column objects above
                    del self.columns._arrays
                except (TypeError, AttributeError), e:
                    # This shouldn't happen as long as self.columns._arrays
                    # is a lazyproperty
                    pass
            elif data is None:
                pass
            else:
                raise TypeError('Table data has incorrect type.')

        if not (isinstance(self._header[0], basestring) and
                self._header[0].rstrip() == self._extension):
            self._header[0] = (self._extension, self._ext_comment)

        # Ensure that the correct EXTNAME is set on the new header if one was
        # created, or that it overrides the existing EXTNAME if different
        if name:
            self.name = name

    @classmethod
    def match_header(cls, header):
        """
        This is an abstract type that implements the shared functionality of
        the ASCII and Binary Table HDU types, which should be used instead of
        this.
        """

        raise NotImplementedError

    @lazyproperty
    def columns(self):
        if self._data_loaded and hasattr(self.data, '_coldefs'):
            return self.data._coldefs
        return ColDefs(self)

    @lazyproperty
    def data(self):
        data = self._get_tbdata()
        data._coldefs = self.columns
        data.formats = self.columns.formats
        # Columns should now just return a reference to the data._coldefs
        del self.columns
        return data

    @lazyproperty
    def _theap(self):
        size = self._header['NAXIS1'] * self._header['NAXIS2']
        return self._header.get('THEAP', size)

    @deprecated('3.0', alternative='the `.columns` attribute')
    def get_coldefs(self):
        """
        Returns the table's column definitions.
        """

        return self.columns

    # TODO: Need to either rename this to update_header, for symmetry with the
    # Image HDUs, or just at some point deprecate it and remove it altogether,
    # since header updates should occur automatically when necessary...
    def update(self):
        """
        Update header keywords to reflect recent changes of columns.
        """

        self._header.set('NAXIS1', self.data.itemsize, after='NAXIS')
        self._header.set('NAXIS2', self.data.shape[0], after='NAXIS1')
        self._header.set('TFIELDS', len(self.columns), after='GCOUNT')

        self._clear_table_keywords()
        self._populate_table_keywords()

    def copy(self):
        """
        Make a copy of the table HDU, both header and data are copied.
        """

        # touch the data, so it's defined (in the case of reading from a
        # FITS file)
        self.data
        return new_table(self.columns, header=self._header,
                         tbtype=self.columns._tbtype)

    def _prewriteto(self, checksum=False, inplace=False):
        if self._data_loaded and self.data is not None:
            self.data._scale_back()
            # check TFIELDS and NAXIS2
            self._header['TFIELDS'] = len(self.data._coldefs)
            self._header['NAXIS2'] = self.data.shape[0]

            # calculate PCOUNT, for variable length tables
            tbsize = self.header['NAXIS1'] * self.header['NAXIS2']
            heapstart = self.header.get('THEAP', tbsize)
            self.data._gap = heapstart - tbsize
            pcount = self.data._heapsize + self.data._gap
            if pcount > 0:
                self.header['PCOUNT'] = pcount

            # update TFORM for variable length columns
            for idx in range(self.data._nfields):
                format = self.data._coldefs.formats[idx]
                if isinstance(format, _FormatP):
                    _max = self.data.field(idx).max
                    format = _FormatP(format.dtype, repeat=format.repeat,
                                      max=_max)
                    self._header['TFORM' + str(idx + 1)] = format.tform
        return super(_TableBaseHDU, self)._prewriteto(checksum, inplace)

    def _verify(self, option='warn'):
        """
        _TableBaseHDU verify method.
        """

        errs = super(_TableBaseHDU, self)._verify(option=option)
        self.req_cards('NAXIS', None, lambda v: (v == 2), 2, option, errs)
        self.req_cards('BITPIX', None, lambda v: (v == 8), 8, option, errs)
        self.req_cards('TFIELDS', 7,
                       lambda v: (_is_int(v) and v >= 0 and v <= 999), 0,
                       option, errs)
        tfields = self._header['TFIELDS']
        for idx in range(tfields):
            self.req_cards('TFORM' + str(idx + 1), None, None, None, option,
                           errs)
        return errs

    def _summary(self):
        """
        Summarize the HDU: name, dimensions, and formats.
        """

        class_name = self.__class__.__name__

        # if data is touched, use data info.
        if self._data_loaded:
            if self.data is None:
                shape, format = (), ''
                nrows = 0
            else:
                nrows = len(self.data)

            ncols = len(self.columns.formats)
            format = self.columns.formats

        # if data is not touched yet, use header info.
        else:
            shape = ()
            nrows = self._header['NAXIS2']
            ncols = self._header['TFIELDS']
            format = ', '.join([self._header['TFORM' + str(j + 1)]
                                for j in range(ncols)])
            format = '[%s]' % format
        dims = "%dR x %dC" % (nrows, ncols)
        ncards = len(self._header)

        return (self.name, class_name, ncards, dims, format)

    def _clear_table_keywords(self):
        """Wipe out any existing table definition keywords from the header."""

        # Go in reverse so as to not confusing indexing while deleting.
        for idx, keyword in enumerate(reversed(self._header.keys())):
            keyword = TDEF_RE.match(keyword)
            try:
                keyword = keyword.group('label')
            except:
                continue                # skip if there is no match
            if (keyword in KEYWORD_NAMES):
                del self._header[idx]

    def _populate_table_keywords(self):
        """Populate the new table definition keywords from the header."""

        cols = self.columns
        append = self._header.append

        for idx, col in enumerate(cols):
            for attr, keyword in zip(KEYWORD_ATTRIBUTES, KEYWORD_NAMES):
                val = getattr(cols, attr + 's')[idx]
                if val:
                    keyword = keyword + str(idx + 1)
                    append((keyword, val))


class TableHDU(_TableBaseHDU):
    """
    FITS ASCII table extension HDU class.
    """

    _extension = 'TABLE'
    _ext_comment = 'ASCII table extension'

    _padding_byte = ' '

    __format_RE = re.compile(
        r'(?P<code>[ADEFIJ])(?P<width>\d+)(?:\.(?P<prec>\d+))?')

    def __init__(self, data=None, header=None, name=None):
        super(TableHDU, self).__init__(data, header, name=name)
        if (self._data_loaded and self.data is not None and
            not isinstance(self.data._coldefs, _ASCIIColDefs)):
            self.data._coldefs = _ASCIIColDefs(self.data._coldefs)

    @classmethod
    def match_header(cls, header):
        card = header.cards[0]
        xtension = card.value
        if isinstance(xtension, basestring):
            xtension = xtension.rstrip()
        return card.keyword == 'XTENSION' and xtension == cls._extension

    def _get_tbdata(self):
        columns = self.columns
        names = [n for idx, n in enumerate(columns.names)
                 if not columns[idx]._phantom]

        # determine if there are duplicate field names and if there
        # are throw an exception
        dup = np.rec.find_duplicate(names)

        if dup:
            raise ValueError("Duplicate field names: %s" % dup)

        itemsize = columns.spans[-1] + columns.starts[-1] - 1
        dtype = {}

        for idx in range(len(columns)):
            data_type = 'S' + str(columns.spans[idx])

            if idx == len(columns) - 1:
                # The last column is padded out to the value of NAXIS1
                if self._header['NAXIS1'] > itemsize:
                    data_type = 'S' + str(columns.spans[idx] + \
                                self._header['NAXIS1'] - itemsize)
            dtype[columns.names[idx]] = (data_type, columns.starts[idx] - 1)

        raw_data = self._get_raw_data(columns._shape, dtype, self._datLoc)
        data = raw_data.view(np.rec.recarray)
        self._init_tbdata(data)
        return data.view(self._data_type)

    def _calculate_datasum(self, blocking):
        """
        Calculate the value for the ``DATASUM`` card in the HDU.
        """

        if self._data_loaded and self.data is not None:
            # We have the data to be used.
            # We need to pad the data to a block length before calculating
            # the datasum.

            if self.size > 0:
                d = np.append(np.fromstring(self.data, dtype='ubyte'),
                              np.fromstring(_pad_length(self.size) * ' ',
                                            dtype='ubyte'))

            cs = self._compute_checksum(np.fromstring(d, dtype='ubyte'),
                                        blocking=blocking)
            return cs
        else:
            # This is the case where the data has not been read from the file
            # yet.  We can handle that in a generic manner so we do it in the
            # base class.  The other possibility is that there is no data at
            # all.  This can also be handled in a gereric manner.
            return super(TableHDU, self)._calculate_datasum(blocking)

    def _verify(self, option='warn'):
        """
        `TableHDU` verify method.
        """

        errs = super(TableHDU, self)._verify(option=option)
        self.req_cards('PCOUNT', None, lambda v: (v == 0), 0, option, errs)
        tfields = self._header['TFIELDS']
        for idx in range(tfields):
            self.req_cards('TBCOL' + str(idx + 1), None, _is_int, None, option,
                           errs)
        return errs


class BinTableHDU(_TableBaseHDU):
    """
    Binary table HDU class.
    """

    _extension = 'BINTABLE'
    _ext_comment = 'binary table extension'

    @classmethod
    def match_header(cls, header):
        card = header.cards[0]
        xtension = card.value
        if isinstance(xtension, basestring):
            xtension = xtension.rstrip()
        return (card.keyword == 'XTENSION' and
                xtension in (cls._extension, 'A3DTABLE'))

    def _calculate_datasum_from_data(self, data, blocking):
        """
        Calculate the value for the ``DATASUM`` card given the input data
        """

        # Check the byte order of the data.  If it is little endian we
        # must swap it before calculating the datasum.
        for i in range(data._nfields):
            coldata = data.field(i)

            if not isinstance(coldata, chararray.chararray):
                if isinstance(coldata, _VLF):
                    for j, d in enumerate(coldata):
                        if not isinstance(d, chararray.chararray):
                            if d.itemsize > 1:
                                if d.dtype.str[0] != '>':
                                    d[:] = d.byteswap()
                                    d.dtype = d.dtype.newbyteorder('>')
                        field = np.rec.recarray.field(data, i)[j:j + 1]
                        if field.dtype.str[0] != '>':
                            field.byteswap(True)
                else:
                    if coldata.itemsize > 1:
                        if data.field(i).dtype.str[0] != '>':
                            data.field(i)[:] = data.field(i).byteswap()
        data.dtype = data.dtype.newbyteorder('>')

        dout = np.fromstring(data, dtype='ubyte')

        for i in range(data._nfields):
            if isinstance(data._coldefs._recformats[i], _FormatP):
                for coldata in data.field(i):
                    if len(coldata) > 0:
                        dout = np.append(dout,
                                         np.fromstring(coldata, dtype='ubyte'))

        cs = self._compute_checksum(dout, blocking=blocking)
        return cs

    def _calculate_datasum(self, blocking):
        """
        Calculate the value for the ``DATASUM`` card in the HDU.
        """

        if self._data_loaded and self.data is not None:
            # We have the data to be used.
            return self._calculate_datasum_from_data(self.data, blocking)
        else:
            # This is the case where the data has not been read from the file
            # yet.  We can handle that in a generic manner so we do it in the
            # base class.  The other possibility is that there is no data at
            # all.  This can also be handled in a generic manner.
            return super(BinTableHDU, self)._calculate_datasum(blocking)

    def _writedata_internal(self, fileobj):
        size = 0

        if self.data is not None:
            size += self._binary_table_byte_swap(fileobj)
            size += self.data.size * self.data.itemsize

        return size

    def _binary_table_byte_swap(self, fileobj):
        swapped = []
        nbytes = 0
        if sys.byteorder == 'little':
            swap_types = ('<', '=')
        else:
            swap_types = ('<',)
        try:
            if not fileobj.simulateonly:
                for idx in range(self.data._nfields):
                    coldata = self.data.field(idx)
                    if isinstance(coldata, chararray.chararray):
                        continue
                    # only swap unswapped
                    # deal with var length table
                    field = np.rec.recarray.field(self.data, idx)
                    if isinstance(coldata, _VLF):
                        for jdx, c in enumerate(coldata):
                            if (not isinstance(c, chararray.chararray) and
                                c.itemsize > 1 and
                                c.dtype.str[0] in swap_types):
                                swapped.append(c)
                            if (field[jdx:jdx + 1].dtype.str[0] in swap_types):
                                swapped.append(field[jdx:jdx + 1])
                    else:
                        if (coldata.itemsize > 1 and
                            self.data.dtype.descr[idx][1][0] in swap_types):
                            swapped.append(field)

                for obj in swapped:
                    obj.byteswap(True)

                fileobj.writearray(self.data)

                # write out the heap of variable length array
                # columns this has to be done after the
                # "regular" data is written (above)
                fileobj.write((self.data._gap * '\0').encode('ascii'))

            nbytes = self.data._gap

            for idx in range(self.data._nfields):
                if isinstance(self.data._coldefs._recformats[idx], _FormatP):
                    field = self.data.field(idx)
                    for jdx in range(len(field)):
                        coldata = field[jdx]
                        if len(coldata) > 0:
                            nbytes = nbytes + coldata.nbytes
                            if not fileobj.simulateonly:
                                fileobj.writearray(coldata)

            self.data._heapsize = nbytes - self.data._gap
        finally:
            for obj in swapped:
                obj.byteswap(True)

        return nbytes

    def _populate_table_keywords(self):
        """Populate the new table definition keywords from the header."""

        cols = self.columns
        append = self._header.append

        for idx, col in enumerate(cols):
            for attr, keyword in zip(KEYWORD_ATTRIBUTES, KEYWORD_NAMES):
                val = getattr(cols, attr + 's')[idx]
                if val:
                    keyword = keyword + str(idx + 1)
                    if attr == 'format':
                        val = cols._recformats[idx]
                        if isinstance(val, _FormatX):
                            val = repr(val._nx) + 'X'
                        elif isinstance(val, _FormatP):
                            val = val.tform
                        else:
                            val = _convert_format(val, reverse=True)
                    append((keyword, val))

    _tdump_file_format = textwrap.dedent("""

        - **datafile:** Each line of the data file represents one row of table
          data.  The data is output one column at a time in column order.  If
          a column contains an array, each element of the column array in the
          current row is output before moving on to the next column.  Each row
          ends with a new line.

          Integer data is output right-justified in a 21-character field
          followed by a blank.  Floating point data is output right justified
          using 'g' format in a 21-character field with 15 digits of
          precision, followed by a blank.  String data that does not contain
          whitespace is output left-justified in a field whose width matches
          the width specified in the ``TFORM`` header parameter for the
          column, followed by a blank.  When the string data contains
          whitespace characters, the string is enclosed in quotation marks
          (``""``).  For the last data element in a row, the trailing blank in
          the field is replaced by a new line character.

          For column data containing variable length arrays ('P' format), the
          array data is preceded by the string ``'VLA_Length= '`` and the
          integer length of the array for that row, left-justified in a
          21-character field, followed by a blank.

          For column data representing a bit field ('X' format), each bit
          value in the field is output right-justified in a 21-character field
          as 1 (for true) or 0 (for false).

        - **cdfile:** Each line of the column definitions file provides the
          definitions for one column in the table.  The line is broken up into
          8, sixteen-character fields.  The first field provides the column
          name (``TTYPEn``).  The second field provides the column format
          (``TFORMn``).  The third field provides the display format
          (``TDISPn``).  The fourth field provides the physical units
          (``TUNITn``).  The fifth field provides the dimensions for a
          multidimensional array (``TDIMn``).  The sixth field provides the
          value that signifies an undefined value (``TNULLn``).  The seventh
          field provides the scale factor (``TSCALn``).  The eighth field
          provides the offset value (``TZEROn``).  A field value of ``""`` is
          used to represent the case where no value is provided.

        - **hfile:** Each line of the header parameters file provides the
          definition of a single HDU header card as represented by the card
          image.
      """)

    def dump(self, datafile=None, cdfile=None, hfile=None, clobber=False):
        """
        Dump the table HDU to a file in ASCII format.  The table may be dumped
        in three separate files, one containing column definitions, one
        containing header parameters, and one for table data.

        Parameters
        ----------
        datafile : file path, file object or file-like object, optional
            Output data file.  The default is the root name of the
            fits file associated with this HDU appended with the
            extension ``.txt``.

        cdfile : file path, file object or file-like object, optional
            Output column definitions file.  The default is `None`, no
            column definitions output is produced.

        hfile : file path, file object or file-like object, optional
            Output header parameters file.  The default is `None`,
            no header parameters output is produced.

        clobber : bool
            Overwrite the output files if they exist.

        Notes
        -----
        The primary use for the `dump` method is to allow viewing and editing
        the table data and parameters in a standard text editor.
        The `load` method can be used to create a new table from the three
        plain text (ASCII) files.
        """

        # TODO: This is looking pretty long and complicated--might be a few
        # places we can break this up into smaller functions

        # check if the output files already exist
        exist = []
        files = [datafile, cdfile, hfile]

        for f in files:
            if isinstance(f, basestring):
                if os.path.exists(f) and os.path.getsize(f) != 0:
                    if clobber:
                        warnings.warn("Overwriting existing file '%s'." % f)
                    else:
                        exist.append(f)

        if exist:
            raise IOError('  '.join(["File '%s' already exists." % f
                                     for f in exist]))

        # Process the data
        self._dump_data(datafile)

        # Process the column definitions
        if cdfile:
            self._dump_coldefs(cdfile)

        # Process the header parameters
        if hfile:
            self._header.tofile(hfile, sep='\n', endcard=False, padding=False)

    dump.__doc__ += _tdump_file_format.replace('\n', '\n        ')

    @deprecated('3.1', alternative=':meth:`dump`')
    def tdump(self, datafile=None, cdfile=None, hfile=None, clobber=False):
        self.dump(datafile, cdfile, hfile, clobber)

    def load(cls, datafile, cdfile=None, hfile=None, replace=False,
             header=None):
        """
        Create a table from the input ASCII files.  The input is from up to
        three separate files, one containing column definitions, one containing
        header parameters, and one containing column data.

        The column definition and header parameters files are not required.
        When absent the column definitions and/or header parameters are taken
        from the header object given in the header argument; otherwise sensible
        defaults are inferred (though this mode is not recommended).

        Parameters
        ----------
        datafile : file path, file object or file-like object
            Input data file containing the table data in ASCII format.

        cdfile : file path, file object, file-like object, optional
            Input column definition file containing the names,
            formats, display formats, physical units, multidimensional
            array dimensions, undefined values, scale factors, and
            offsets associated with the columns in the table.  If
            `None`, the column definitions are taken from the current
            values in this object.

        hfile : file path, file object, file-like object, optional
            Input parameter definition file containing the header
            parameter definitions to be associated with the table.  If
            `None`, the header parameter definitions are taken from
            the current values in this objects header.

        replace : bool
            When `True`, indicates that the entire header should be
            replaced with the contents of the ASCII file instead of
            just updating the current header.

        header : Header object
            When the cdfile and hfile are missing, use this Header object in
            the creation of the new table and HDU.  Otherwise this Header
            supercedes the keywords from hfile, which is only used to update
            values not present in this Header, unless replace=True in which
            this Header's values are completely replaced with the values from
            hfile.

        Notes
        -----
        The primary use for the `load` method is to allow the input of ASCII
        data that was edited in a standard text editor of the table data and
        parameters.  The `dump` method can be used to create the initial ASCII
        files.
        """

        # Process the parameter file
        if header is None:
            header = Header()

        if hfile:
            if replace:
                header = Header.fromtextfile(hfile)
            else:
                header.extend(Header.fromtextfile(hfile), update=True,
                              update_first=True)

        coldefs = None
        # Process the column definitions file
        if cdfile:
            coldefs = cls._load_coldefs(cdfile)

        # Process the data file
        data = cls._load_data(datafile, coldefs)
        if coldefs is None:
            coldefs = ColDefs(data)

        # Create a new HDU using the supplied header and data
        hdu = cls(data=data, header=header)
        hdu.columns = coldefs
        return hdu
    load.__doc__ += _tdump_file_format.replace('\n', '\n        ')
    load = classmethod(load)
    # Have to create a classmethod from this here instead of as a decorator;
    # otherwise we can't update __doc__

    @deprecated('3.1', alternative=':meth:`load`')
    @classmethod
    def tcreate(cls, datafile, cdfile=None, hfile=None, replace=False,
                header=None):
        return cls.load(datafile, cdfile, hfile, replace, header)

    def _dump_data(self, fileobj):
        """
        Write the table data in the ASCII format read by BinTableHDU.load()
        to fileobj.
        """

        if not fileobj and self._file:
            root, ext = os.path.splitext(self._file.name)
            fileobj = root + '.txt'

        close_file = False

        if isinstance(fileobj, basestring):
            fileobj = open(fileobj, 'w')
            close_file = True

        linewriter = csv.writer(fileobj, dialect=FITSTableDumpDialect)

        # Process each row of the table and output one row at a time
        def format_value(val, format):
            if format[0] == 'S':
                itemsize = int(format[1:])
                return '%-*s' % (itemsize, val)
            elif format in np.typecodes['AllInteger']:
                # output integer
                return '%21d' % val
            elif format in np.typecodes['Complex']:
                return '%21.15g+%.15gj' % (val.real, val.imag)
            elif format in np.typecodes['Float']:
                # output floating point
                return '%#21.15g' % val

        for row in self.data:
            line = []   # the line for this row of the table

            # Process each column of the row.
            for column in self.columns:
                vla_format = None   # format of data in a variable length array
                                    # where None means it is not a VLA
                format = _convert_format(column.format)

                if isinstance(format, _FormatP):
                    # P format means this is a variable length array so output
                    # the length of the array for this row and set the format
                    # for the VLA data
                    line.append('VLA_Length=')
                    line.append('%-21d' % len(row[column.name]))
                    repeat, dtype, option = _parse_tformat(column.format)
                    vla_format = FITS2NUMPY[option[0]][0]

                if vla_format:
                    # Output the data for each element in the array
                    for val in row[column.name].flat:
                        line.append(format_value(val, vla_format))
                else:
                    # The column data is a single element
                    dtype = self.data.dtype.fields[column.name][0]
                    array_format = dtype.char
                    if array_format == 'S':
                        array_format += str(dtype.itemsize)
                    line.append(format_value(row[column.name], array_format))
            linewriter.writerow(line)
        if close_file:
            fileobj.close()

    def _dump_coldefs(self, fileobj):
        """
        Write the column definition parameters in the ASCII format read by
        BinTableHDU.load() to fileobj.
        """

        close_file = False

        if isinstance(fileobj, basestring):
            fileobj = open(fileobj, 'w')
            close_file = True

        # Process each column of the table and output the result to the
        # file one at a time
        for column in self.columns:
            line = [column.name, column.format]
            attrs = ['disp', 'unit', 'dim', 'null', 'bscale', 'bzero']
            line += ['%-16s' % (value if value else '""')
                     for value in (getattr(column, attr) for attr in attrs)]
            fileobj.write(' '.join(line))
            fileobj.write('\n')

        if close_file:
            fileobj.close()

    @classmethod
    def _load_data(cls, fileobj, coldefs=None):
        """
        Read the table data from the ASCII file output by BinTableHDU.dump().
        """

        close_file = False

        if isinstance(fileobj, basestring):
            fileobj = open(fileobj, 'r')
            close_file = True

        initialpos = fileobj.tell()  # We'll be returning here later
        linereader = csv.reader(fileobj, dialect=FITSTableDumpDialect)

        # First we need to do some preprocessing on the file to find out how
        # much memory we'll need to reserve for the table.  This is necessary
        # even if we already have the coldefs in order to determine how many
        # rows to reserve memory for
        vla_lengths = []
        recformats = []
        names = []
        nrows = 0
        if coldefs is not None:
            recformats = coldefs._recformats
            names = coldefs.names

        def update_recformats(value, idx):
            fitsformat = _scalar_to_format(value)
            recformat = _convert_format(fitsformat)
            if idx >= len(recformats):
                recformats.append(recformat)
            else:
                if _cmp_recformats(recformats[idx], recformat) < 0:
                    recformats[idx] = recformat

        # TODO: The handling of VLAs could probably be simplified a bit
        for row in linereader:
            nrows += 1
            if coldefs is not None:
                continue
            col = 0
            idx = 0
            while idx < len(row):
                if row[idx] == 'VLA_Length=':
                    if col < len(vla_lengths):
                        vla_length = vla_lengths[col]
                    else:
                        vla_length = int(row[idx + 1])
                        vla_lengths.append(vla_length)
                    idx += 2
                    while vla_length:
                        update_recformats(row[idx], col)
                        vla_length -= 1
                        idx += 1
                    col += 1
                else:
                    if col >= len(vla_lengths):
                        vla_lengths.append(None)
                    update_recformats(row[idx], col)
                    col += 1
                    idx += 1

        # Update the recformats for any VLAs
        for idx, length in enumerate(vla_lengths):
            if length is not None:
                recformats[idx] = str(length) + recformats[idx]

        dtype = np.rec.format_parser(recformats, names, None).dtype

        # TODO: In the future maybe enable loading a bit at a time so that we
        # can convert from this format to an actual FITS file on disk without
        # needing enough physical memory to hold the entire thing at once;
        # new_table() could use a similar feature.
        hdu = new_table(np.recarray(shape=1, dtype=dtype), nrows=nrows,
                        fill=True)
        data = hdu.data
        for idx, length in enumerate(vla_lengths):
            if length is not None:
                arr = data.columns._arrays[idx]
                dt = recformats[idx][len(str(length)):]
                recformats[idx] = _FormatP(dt, max=length)
                data.columns._recformats[idx] = recformats[idx]
                data._convert[idx] = _makep(arr, arr, recformats[idx])

        # Jump back to the start of the data and create a new line reader
        fileobj.seek(initialpos)
        linereader = csv.reader(fileobj, dialect=FITSTableDumpDialect)
        for row, line in enumerate(linereader):
            col = 0
            idx = 0
            while idx < len(line):
                if line[idx] == 'VLA_Length=':
                    vla_len = vla_lengths[col]
                    idx += 2
                    data[row][col][:] = line[idx:idx + vla_len]
                    idx += vla_len
                else:
                    # TODO: This won't work for complex-valued types; fix this
                    # Kind of silly special handling for bools
                    val = line[idx]
                    if recformats[col] == FITS2NUMPY['L']:
                        val = bool(int(val))
                    elif recformats[col] == FITS2NUMPY['M']:
                        # For some reason, in arrays/fields where numpy expects
                        # a complex it's not happy to take a string
                        # representation (though it's happy to do that in other
                        # contexts), so we have to convert the string
                        # representation for it:
                        val = complex(val)
                    data[row][col] = val
                    idx += 1
                col += 1

        if close_file:
            fileobj.close()

        return data

    @classmethod
    def _load_coldefs(cls, fileobj):
        """
        Read the table column definitions from the ASCII file output by
        BinTableHDU.dump().
        """

        close_file = False

        if isinstance(fileobj, basestring):
            fileobj = open(fileobj, 'r')
            close_file = True

        columns = []

        for line in fileobj:
            words = line[:-1].split()
            kwargs = {}
            for key in ['name', 'format', 'disp', 'unit', 'dim']:
                kwargs[key] = words.pop(0).replace('""', '')

            for key in ['null', 'bscale', 'bzero']:
                word = words.pop(0).replace('""', '')
                if word:
                    word = _str_to_num(word)
                kwargs[key] = word
            columns.append(Column(**kwargs))

        if close_file:
            fileobj.close()

        return ColDefs(columns)


# TODO: Allow tbtype to be either a string or a class; perhaps eventually
# replace this with separate functions for creating tables (possibly in the
# form of a classmethod)  See ticket #60
def new_table(input, header=None, nrows=0, fill=False, tbtype='BinTableHDU'):
    """
    Create a new table from the input column definitions.

    Warning: Creating a new table using this method creates an in-memory *copy*
    of all the column arrays in the input.  This is because if they are
    separate arrays they must be combined into a single contiguous array.

    If the column data is already in a single contiguous array (such as an
    existing record array) it may be better to create a BinTableHDU instance
    directly.  See the PyFITS documentation for more details.

    Parameters
    ----------
    input : sequence of Column or ColDefs objects
        The data to create a table from.

    header : Header instance
        Header to be used to populate the non-required keywords.

    nrows : int
        Number of rows in the new table.

    fill : bool
        If `True`, will fill all cells with zeros or blanks.  If
        `False`, copy the data from input, undefined cells will still
        be filled with zeros/blanks.

    tbtype : str
        Table type to be created ("BinTableHDU" or "TableHDU").
    """

    # construct a table HDU
    # TODO: Something needs to be done about this as part of #60....
    hdu = eval(tbtype)(header=header)

    if isinstance(input, ColDefs):
        # NOTE: This previously raised an error if the tbtype didn't match the
        # tbtype of the input ColDefs. This should no longer be necessary, but
        # just beware.
        columns = hdu.columns = ColDefs(input)
    elif isinstance(input, FITS_rec):  # input is a FITS_rec
        # Create a new ColDefs object from the input FITS_rec's ColDefs
        # object and assign it to the ColDefs attribute of the new hdu.
        columns = hdu.columns = ColDefs(input._coldefs, tbtype)
    else:  # input is a list of Columns or possibly a recarray
        # Create a new ColDefs object from the input list of Columns and
        # assign it to the ColDefs attribute of the new hdu.
        columns = hdu.columns = ColDefs(input, tbtype)

    # read the delayed data
    for idx in range(len(columns)):
        arr = columns._arrays[idx]
        if isinstance(arr, Delayed):
            if arr.hdu.data is None:
                columns._arrays[idx] = None
            else:
                columns._arrays[idx] = np.rec.recarray.field(arr.hdu.data,
                                                             arr.field)

    # use the largest column shape as the shape of the record
    if nrows == 0:
        for arr in columns._arrays:
            if (arr is not None):
                dim = arr.shape[0]
            else:
                dim = 0
            if dim > nrows:
                nrows = dim

    if tbtype == 'TableHDU':
        columns = hdu.columns = _ASCIIColDefs(hdu.columns)
        _itemsize = columns.spans[-1] + columns.starts[-1] - 1
        dtype = {}

        for j in range(len(columns)):
            data_type = 'S' + str(columns.spans[j])
            dtype[columns.names[j]] = (data_type, columns.starts[j] - 1)

        hdu.data = np.rec.array((' ' * _itemsize * nrows).encode('ascii'),
                                dtype=dtype, shape=nrows).view(FITS_rec)
        hdu.data.setflags(write=True)
    else:
        formats = ','.join(columns._recformats)
        hdu.data = np.rec.array(None, formats=formats,
                                names=columns.names,
                                shape=nrows).view(FITS_rec)

    hdu.data._coldefs = hdu.columns
    hdu.data.formats = hdu.columns.formats

    # Populate data to the new table from the ndarrays in the input ColDefs
    # object.
    for idx in range(len(columns)):
        # For each column in the ColDef object, determine the number
        # of rows in that column.  This will be either the number of
        # rows in the ndarray associated with the column, or the
        # number of rows given in the call to this function, which
        # ever is smaller.  If the input FILL argument is true, the
        # number of rows is set to zero so that no data is copied from
        # the original input data.
        arr = columns._arrays[idx]
        recformat = columns._recformats[idx]

        if arr is None:
            size = 0
        else:
            size = len(arr)

        n = min(size, nrows)
        if fill:
            n = 0

        # Get any scale factors from the FITS_rec
        scale, zero, bscale, bzero, dim = hdu.data._get_scale_factors(idx)[3:]

        field = np.rec.recarray.field(hdu.data, idx)

        if n > 0:
            # Only copy data if there is input data to copy
            # Copy all of the data from the input ColDefs object for this
            # column to the new FITS_rec data array for this column.
            if isinstance(recformat, _FormatX):
                # Data is a bit array
                if arr[:n].shape[-1] == recformat._nx:
                    _wrapx(arr[:n], field[:n], recformat._nx)
                else:
                    # from a table parent data, just pass it
                    field[:n] = arr[:n]
            elif isinstance(recformat, _FormatP):
                hdu.data._convert[idx] = _makep(arr[:n], field, recformat,
                                                nrows=nrows)
            elif recformat[-2:] == FITS2NUMPY['L'] and arr.dtype == bool:
                # column is boolean
                field[:n] = np.where(arr == False, ord('F'), ord('T'))
            else:
                if tbtype == 'TableHDU':
                    # string no need to convert,
                    if isinstance(arr, chararray.chararray):
                        field[:n] = arr[:n]
                    else:
                        hdu.data._convert[idx] = \
                                np.zeros(nrows, dtype=arr.dtype)
                        if scale or zero:
                            arr = arr.copy()
                        if scale:
                            arr *= bscale
                        if zero:
                            arr += bzero
                        hdu.data._convert[idx][:n] = arr[:n]
                else:
                    outarr = field[:n]
                    inarr = arr[:n]
                    if inarr.shape != outarr.shape:
                        if inarr.dtype != outarr.dtype:
                            inarr = inarr.view(outarr.dtype)

                        # This is a special case to handle input arrays with
                        # non-trivial TDIMn.
                        # By design each row of the outarray is 1-D, while each
                        # row of the input array may be n-D
                        if outarr.ndim > 1:
                            # The normal case where the first dimension is the
                            # rows
                            inarr_rowsize = inarr[0].size
                            inarr = inarr.reshape((n, inarr_rowsize))
                            outarr[:,:inarr_rowsize] = inarr
                        else:
                            # Special case for strings where the out array only
                            # has one dimension (the second dimension is rolled
                            # up into the strings
                            outarr[:n] = inarr.ravel()
                    else:
                        field[:n] = arr[:n]

        if n < nrows:
            # If there are additional rows in the new table that were not
            # copied from the input ColDefs object, initialize the new data
            if tbtype == 'BinTableHDU':
                if isinstance(field, np.ndarray):
                    field[n:] = -bzero / bscale
                else:
                    field[n:] = ''
            else:
                field[n:] = ' ' * hdu.data._coldefs.spans[idx]

    # Update the HDU header to match the data
    hdu.update()

    # Make the ndarrays in the Column objects of the ColDefs object of the HDU
    # reference the same ndarray as the HDU's FITS_rec object.
    for idx in range(len(columns)):
        hdu.columns[idx].array = hdu.data.field(idx)

    # Delete the _arrays attribute so that it is recreated to point to the
    # new data placed in the column objects above
    del hdu.columns._arrays

    return hdu



from __future__ import division, with_statement  # confidence high

import warnings

import numpy as np

import pyfits
from pyfits.tests import PyfitsTestCase

from nose.tools import assert_equal, assert_raises, assert_true, assert_false


class TestChecksumFunctions(PyfitsTestCase):

    def setup(self):
        super(TestChecksumFunctions, self).setup()
        self._oldfilters = warnings.filters[:]
        warnings.filterwarnings(
            'error',
             message='Checksum verification failed')
        warnings.filterwarnings(
            'error',
            message='Datasum verification failed')

    def teardown(self):
        super(TestChecksumFunctions, self).teardown()
        warnings.filters = self._oldfilters

    def test_sample_file(self):
        hdul = pyfits.open(self.data('checksum.fits'), checksum=True)
        hdul.close()

    def test_image_create(self):
        n = np.arange(100)
        hdu = pyfits.PrimaryHDU(n)
        hdu.writeto(self.temp('tmp.fits'), clobber=True, checksum=True)
        hdul = pyfits.open(self.temp('tmp.fits'), checksum=True)
        hdul.close()

    def test_nonstandard_checksum(self):
        hdu = pyfits.PrimaryHDU(np.arange(10.0 ** 6))
        hdu.writeto(self.temp('tmp.fits'), clobber=True,
                    checksum='nonstandard')
        del hdu
        hdul = pyfits.open(self.temp('tmp.fits'), checksum='nonstandard')

    def test_scaled_data(self):
        hdul = pyfits.open(self.data('scale.fits'))
        orig_data = hdul[0].data.copy()
        hdul[0].scale('int16', 'old')
        hdul.writeto(self.temp('tmp.fits'), clobber=True, checksum=True)
        hdul1 = pyfits.open(self.temp('tmp.fits'), checksum=True)
        assert_true((hdul1[0].data == orig_data).all())
        hdul.close()
        hdul1.close()

    def test_uint16_data(self):
        hdul = pyfits.open(self.data('o4sp040b0_raw.fits'), uint=True)
        hdul.writeto(self.temp('tmp.fits'), clobber=True, checksum=True)
        hdul1 = pyfits.open(self.temp('tmp.fits'), uint=True, checksum=True)
        hdul.close()
        hdul1.close()

    def test_groups_hdu_data(self):
        imdata = np.arange(100.0)
        imdata.shape = (10, 1, 1, 2, 5)
        pdata1 = np.arange(10) + 0.1
        pdata2 = 42
        x = pyfits.hdu.groups.GroupData(imdata, parnames=['abc', 'xyz'],
                                        pardata=[pdata1, pdata2], bitpix=-32)
        hdu = pyfits.GroupsHDU(x)
        hdu.writeto(self.temp('tmp.fits'), clobber=True, checksum=True)
        hdul1 = pyfits.open(self.temp('tmp.fits'), checksum=True)
        hdul1.close()

    def test_binary_table_data(self):
        a1 = np.array(['NGC1001', 'NGC1002', 'NGC1003'])
        a2 = np.array([11.1, 12.3, 15.2])
        col1 = pyfits.Column(name='target', format='20A', array=a1)
        col2 = pyfits.Column(name='V_mag', format='E', array=a2)
        cols = pyfits.ColDefs([col1, col2])
        tbhdu = pyfits.new_table(cols)
        tbhdu.writeto(self.temp('tmp.fits'), clobber=True, checksum=True)
        hdul = pyfits.open(self.temp('tmp.fits'), checksum=True)
        hdul.close()

    def test_variable_length_table_data(self):
        c1 = pyfits.Column(name='var', format='PJ()',
            array=np.array([[45.0, 56], np.array([11, 12, 13])], 'O'))
        c2 = pyfits.Column(name='xyz', format='2I', array=[[11, 3], [12, 4]])
        tbhdu = pyfits.new_table([c1, c2])
        tbhdu.writeto(self.temp('tmp.fits'), clobber=True, checksum=True)
        hdul = pyfits.open(self.temp('tmp.fits'), checksum=True)
        hdul.close()

    def test_ascii_table_data(self):
        a1 = np.array(['abc', 'def'])
        r1 = np.array([11.0, 12.0])
        c1 = pyfits.Column(name='abc', format='A3', array=a1)
        c2 = pyfits.Column(name='def', format='E', array=r1, bscale=2.3,
                           bzero=0.6)
        c3 = pyfits.Column(name='t1', format='I', array=[91, 92, 93])
        x = pyfits.ColDefs([c1, c2, c3], tbtype='TableHDU')
        hdu = pyfits.new_table(x, tbtype='TableHDU')
        hdu.writeto(self.temp('tmp.fits'), clobber=True, checksum=True)
        hdul = pyfits.open(self.temp('tmp.fits'), checksum=True)
        hdul.close()

    def test_compressed_image_data(self):
        hdul = pyfits.open(self.data('comp.fits'))
        hdul.writeto(self.temp('tmp.fits'), clobber=True, checksum=True)
        hdul1 = pyfits.open(self.temp('tmp.fits'), checksum=True)
        hdul1.close()
        hdul.close()

    def test_compressed_image_data_int16(self):
        n = np.arange(100, dtype='int16')
        hdu = pyfits.ImageHDU(n)
        comp_hdu = pyfits.CompImageHDU(hdu.data, hdu.header)
        comp_hdu.writeto(self.temp('tmp.fits'), checksum=True)
        hdul = pyfits.open(self.temp('tmp.fits'), checksum=True)
        hdul.close()

    def test_compressed_image_data_float32(self):
        n = np.arange(100, dtype='float32')
        comp_hdu = pyfits.CompImageHDU(n)
        comp_hdu.writeto(self.temp('tmp.fits'), checksum=True)
        hdul = pyfits.open(self.temp('tmp.fits'), checksum=True)
        hdul.close()

    def test_open_with_no_keywords(self):
        hdul=pyfits.open(self.data('arange.fits'), checksum=True)
        hdul.close()

    def test_append(self):
        hdul = pyfits.open(self.data('tb.fits'))
        hdul.writeto(self.temp('tmp.fits'), clobber=True)
        n = np.arange(100)
        pyfits.append(self.temp('tmp.fits'), n, checksum=True)
        hdul.close()
        hdul = pyfits.open(self.temp('tmp.fits'), checksum=True)
        assert_equal(hdul[0]._checksum, None)
        hdul.close()

    def test_writeto_convenience(self):
        n = np.arange(100)
        pyfits.writeto(self.temp('tmp.fits'), n, clobber=True, checksum=True)
        hdul = pyfits.open(self.temp('tmp.fits'), checksum=True)

        assert_true(hasattr(hdul[0], '_datasum') and hdul[0]._datasum,
                    msg='Missing DATASUM keyword')

        assert_true(hasattr(hdul[0], '_checksum') and hdul[0]._checksum,
                    msg='Missing CHECKSUM keyword')

        assert_true(hasattr(hdul[0], '_datasum_comment') and
                    hdul[0]._datasum_comment,
                    msg='Missing DATASUM Card comment')

        assert_true(hasattr(hdul[0], '_checksum_comment') and
                    hdul[0]._checksum_comment,
                    msg='Missing CHECKSUM Card comment')

        hdul.close()

    def test_hdu_writeto(self):
        n = np.arange(100, dtype='int16')
        hdu = pyfits.ImageHDU(n)
        hdu.writeto(self.temp('tmp.fits'), checksum=True)
        hdul = pyfits.open(self.temp('tmp.fits'), checksum=True)

        assert_true(hasattr(hdul[0], '_datasum') and hdul[0]._datasum,
                    msg='Missing DATASUM keyword')

        assert_true(hasattr(hdul[0], '_checksum') and hdul[0]._checksum,
                    msg='Missing CHECKSUM keyword')

        assert_true(hasattr(hdul[0], '_datasum_comment') and
                    hdul[0]._datasum_comment,
                    msg='Missing DATASUM Card comment')

        assert_true(hasattr(hdul[0], '_checksum_comment') and
                    hdul[0]._checksum_comment,
                    msg='Missing CHECKSUM Card comment')

        hdul.close()

    def test_datasum_only(self):
        n = np.arange(100, dtype='int16')
        hdu = pyfits.ImageHDU(n)
        hdu.writeto(self.temp('tmp.fits'), clobber=True, checksum='datasum')
        hdul = pyfits.open(self.temp('tmp.fits'), checksum=True)

        assert_true(hasattr(hdul[0], '_datasum') and hdul[0]._datasum,
                    msg='Missing DATASUM keyword')

        assert_true(hasattr(hdul[0], '_checksum') and not hdul[0]._checksum,
                    msg='CHECKSUM keyword should be blank')

        assert_true(hasattr(hdul[0], '_datasum_comment') and
                    hdul[0]._datasum_comment,
                    msg='Missing DATASUM Card comment')

        assert_true(hasattr(hdul[0], '_checksum_comment') and
                    not hdul[0]._checksum_comment,
                    msg='CHECKSUM Card comment should be blank')

        hdul.close()

    def test_open_update_mode_preserve_checksum(self):
        """
        Regression test for #148 where checksums are being removed from headers
        when a file is opened in update mode, even though no changes were made
        to the file.
        """

        self.copy_file('checksum.fits')

        with pyfits.open(self.temp('checksum.fits')) as hdul:
            data = hdul[1].data.copy()

        hdul = pyfits.open(self.temp('checksum.fits'), mode='update')
        hdul.close()

        with pyfits.open(self.temp('checksum.fits')) as hdul:
            assert_true('CHECKSUM' in hdul[1].header)
            assert_true('DATASUM' in hdul[1].header)
            assert_true((data == hdul[1].data).all())

    def test_open_update_mode_update_checksum(self):
        """
        Regression test for #148, part 2.  This ensures that if a file contains
        a checksum, the checksum is updated when changes are saved to the file,
        even if the file was opened with the default of checksum=False.

        An existing checksum and/or datasum are only stripped if the file is
        opened with checksum='remove'.
        """

        self.copy_file('checksum.fits')
        with pyfits.open(self.temp('checksum.fits')) as hdul:
            header = hdul[1].header.copy()
            data = hdul[1].data.copy()

        with pyfits.open(self.temp('checksum.fits'), mode='update') as hdul:
            hdul[1].header['FOO'] = 'BAR'
            hdul[1].data[0]['TIME'] = 42

        with pyfits.open(self.temp('checksum.fits')) as hdul:
            header2 = hdul[1].header
            data2 = hdul[1].data
            assert_equal(header2[:-3], header[:-2])
            assert_true('CHECKSUM' in header2)
            assert_true('DATASUM' in header2)
            assert_equal(header2['FOO'], 'BAR')
            assert_true((data2['TIME'][1:] == data['TIME'][1:]).all())
            assert_equal(data2['TIME'][0], 42)

        with pyfits.open(self.temp('checksum.fits'), mode='update',
                         checksum='remove') as hdul:
            pass

        with pyfits.open(self.temp('checksum.fits')) as hdul:
            header2 = hdul[1].header
            data2 = hdul[1].data
            assert_equal(header2[:-1], header[:-2])
            assert_false('CHECKSUM' in header2)
            assert_false('DATASUM' in header2)
            assert_equal(header2['FOO'], 'BAR')
            assert_true((data2['TIME'][1:] == data['TIME'][1:]).all())
            assert_equal(data2['TIME'][0], 42)
